import re

from bzt.jmx.base import JMX
from bzt.jmx.tools import ProtocolHandler
from lxml import etree


class RTEProtocolHandler(ProtocolHandler):
    def get_toplevel_elements(self, scenario):
        if hasattr(scenario, 'data') and "protocol" in scenario.data and scenario.data['protocol'] == "rte":
            return self._gen_server_config(scenario.data)
        return []

    def _gen_server_config(self, config_data):
        config = etree.Element("ConfigTestElement",
                            guiclass="com.blazemeter.jmeter.rte.sampler.gui.RTEConfigGui",
                            testclass="ConfigTestElement",
                            testname="Server Config")

        if "rte-server" in config_data:
            server = config_data.get("rte-server").split(":")
            config.append(JMX._string_prop("RTEConnectionConfig.server", server[0]))
            if len(server) > 1:
                config.append(JMX._string_prop("RTEConnectionConfig.port", server[1]))

        if 'rte-protocol' in config_data:
            config.append(JMX._string_prop("RTEConnectionConfig.protocol", config_data['rte-protocol']))

        if 'rte-config' in config_data:
            rte_config = config_data.get('rte-config')
            if 'ssl' in rte_config:
                config.append(JMX._string_prop("RTEConnectionConfig.sslType", rte_config['ssl']))
            if 'terminal-type' in rte_config:
                config.append(JMX._string_prop("RTEConnectionConfig.terminalType", rte_config['terminal-type']))
            if 'connection-timeout' in rte_config:
                connection_timeout = self.safe_time(rte_config['connection-timeout'])
                config.append(JMX._string_prop("RTEConnectionConfig.connectTimeout", connection_timeout))

        return [config, etree.Element("hashTree")]

    def get_sampler_pair(self, request):
        action = etree.Element("com.blazemeter.jmeter.rte.sampler.RTESampler",
                               guiclass="com.blazemeter.jmeter.rte.sampler.gui.RTESamplerGui",
                               testclass="com.blazemeter.jmeter.rte.sampler.",
                               testname=request.label)

        action.append(JMX._string_prop("RTESampler.action", request.config.get('rte-action')))
        if "wait-for" in request.config:
            self._gen_wait(action, request.config.get("wait-for"))

        if 'fields' in request.config:
            inputs = JMX._collection_prop("Inputs.inputs")
            for _input in request.config["fields"]:
                inputs.append(self._gen_input_prop(_input))
            elem_props = JMX._element_prop("Inputs.inputs", "com.blazemeter.jmeter.rte.sampler.Inputs")
            elem_props.append(inputs)
            action.append(elem_props)

        if 'attention-key' in request.config:
            action.append(JMX._string_prop("RTESampler.attentionKey", request.config.get('attention-key')))

        children = etree.Element("hashTree")
        if 'extract-cursor-position' in request.config and request.config.get('extract-cursor-position'):
            element = request.config.get('extract-cursor-position')

            extractor = self._gen_rte_extractor(element['varname'], "CURSOR_POSITION")
            children.append(extractor)
            children.append(etree.Element("hashTree"))  # Is this an RTE extension bug? doesn't seem necesary,
            # but won't work without it.

        if 'extract-field-position' in request.config and request.config.get('extract-field-position'):
            element = request.config.get('extract-field-position')
            position = element['base-position'].strip("()")
            row, column = position.split(",")
            offset = element['field-offset']
            extractor = self._gen_rte_extractor(element['varname'], "NEXT_FIELD_POSITION", row, column, offset)
            children.append(extractor)
            children.append(etree.Element("hashTree"))

        return action, children

    @staticmethod
    def _gen_rte_extractor(varname, type, row="", column="", offset=""):
        extractor = etree.Element("com.blazemeter.jmeter.rte.extractor.RTEExtractor",
                                  guiclass="com.blazemeter.jmeter.rte.extractor.RTEExtractorGui",
                                  testclass="com.blazemeter.jmeter.rte.extractor.RTEExtractor",
                                  testname="Cursor extractor")
        extractor.append(JMX._string_prop("RTEExtractor.variablePrefix", varname))
        extractor.append(JMX._string_prop("RTEExtractor.positionType", type))
        extractor.append(JMX._string_prop("RTEExtractor.row", row))
        extractor.append(JMX._string_prop("RTEExtractor.offset", offset))
        extractor.append(JMX._string_prop("RTEExtractor.column", column))
        return extractor

    @staticmethod
    def _gen_input_prop(_input):
        row, col, label, text, tabs = RTEProtocolHandler._handle_input_notation(_input)
        if row and col:
            prop = JMX._element_prop("elementProp", "com.blazemeter.jmeter.rte.sampler.CoordInputRowGUI")
            prop.append(JMX._string_prop("CoordInputRowGUI.row", row))
            prop.append(JMX._string_prop("CoordInputRowGUI.column", col))
        elif label:
            prop = JMX._element_prop("elementProp", "com.blazemeter.jmeter.rte.sampler.LabelInputRowGUI")
            prop.append(JMX._string_prop("LabelInputRowGUI.label", label))
        elif tabs is not None:
            prop = JMX._element_prop("elementProp", "com.blazemeter.jmeter.rte.sampler.TabulatorInputRowGui")
            prop.append(JMX._string_prop("TabulatorInputOffsetGUI.column", tabs))
        prop.append(JMX._string_prop("Input.input", text))
        return prop

    @staticmethod
    def _handle_input_notation(_input):
        row = _input.get('row', None)
        col = _input.get('column', None)
        label = _input.get('label', None)
        text = _input.get('text', None)
        tabs = _input.get('tabs', None)

        # Handle shorthand declarations
        for key, value in _input.items():
            if re.match(r'\(\d+,\d+\)', key):
                row, col = re.findall(r'\d+', key)
                text = _input[key]

            elif re.match(r'\(.+?\)', key):
                label = re.findall("\((.+?)\)", key)[0]
                text = _input[key]
        return row, col, label, text, tabs

    def _gen_wait(self, action, wait_for):
        action.append(JMX._bool_prop("RTESampler.waitSync", 'sync' in wait_for))
        if 'sync' in wait_for:
            sync_timeout = self.safe_time(wait_for['sync'])
            action.append(JMX._string_prop("RTESampler.waitSyncTimeout", sync_timeout))

        action.append(JMX._bool_prop("RTESampler.waitCursor", 'cursor' in wait_for))
        if 'cursor' in wait_for:
            cursor = wait_for['cursor']
            if type(cursor) is str:
                row, column = cursor.split(',')
            else:
                row, column = cursor['position'].split(',')
            action.append(JMX._string_prop("RTESampler.waitCursorRow", row))
            action.append(JMX._string_prop("RTESampler.waitCursorColumn", column))
            if 'timeout' in cursor:
                cursor_timeout = self.safe_time(cursor['timeout'])
                action.append(JMX._string_prop("RTESampler.waitCursorTimeout", cursor_timeout))

        action.append(JMX._bool_prop("RTESampler.waitSilent", 'silent' in wait_for))
        if 'silent' in wait_for:
            silent = wait_for['silent']
            if type(silent) is str:
                stable_time = self.safe_time(silent)
            else:
                stable_time = self.safe_time(silent['stable-time'])

            action.append(JMX._string_prop("RTESampler.waitSilentTime", stable_time))
            if 'timeout' in silent:
                silent_timeout = self.safe_time(silent['timeout'])
                action.append(JMX._string_prop("RTESampler.waitSilentTimeout", silent_timeout))

        action.append(JMX._bool_prop("RTESampler.waitText", 'text' in wait_for))
        if 'text' in wait_for:
            text = wait_for['text']
            if type(text) is str:
                regex = text
            else:
                regex = text['regex']
            action.append(JMX._string_prop("RTESampler.waitTextRegex", regex))

            if 'timeout' in text:
                timeout = self.safe_time(text['timeout'])
                action.append(JMX._string_prop("RTESampler.waitTextTimeout", timeout))

            if 'search-area' in text:
                top, left, bot, right = re.findall(r'\d+', text['search-area'])
                action.append(JMX._string_prop("RTESampler.waitTextAreaTop", top))
                action.append(JMX._string_prop("RTESampler.waitTextAreaLeft", left))
                action.append(JMX._string_prop("RTESampler.waitTextAreaBottom", bot))
                action.append(JMX._string_prop("RTESampler.waitTextAreaRight", right))



"""
Copyright 2018 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import time
from os import listdir
from os.path import join

from bzt import ToolError
from bzt.modules import SubprocessedExecutor
from bzt.modules.functional import FuncSamplesReader
from bzt.modules.jmeter import JTLReader
from bzt.utils import get_full_path, shell_exec, TclLibrary, JavaVM, BetterDict, get_assembled_value
from .tools import Hamcrest, Json, TaurusJavaHelper, JavaC, JUnitJupiterApi, JUnitJupiterEngine
from .tools import JUnitPlatformCommons, JUnitPlatformLauncher, JUnitPlatformEngine, JUnitPlatformRunner
from .tools import JUnitPlatformSuiteApi, JUnitVintageEngine, ApiGuardian, JUnit, OpenTest4j, TestNG


class JavaTestRunner(SubprocessedExecutor):
    """
    Allows to test java and jar files
    :type script: str
    """
    def __init__(self):
        super(JavaTestRunner, self).__init__()
        self.working_dir = os.getcwd()
        self.target_java = "1.8"
        self.props_file = None
        self.class_path = []
        self._tools = []
        self._java_scripts = []
        self._full_install = True
        self.tool = None
        self.report_file_suffix = ".ldjson"

    def install_required_tools(self):
        self._add_jar_tool(Hamcrest, config=self.settings.get("hamcrest-core"))
        self._add_jar_tool(Json, config=self.settings.get("json-jar"))
        self._add_jar_tool(TaurusJavaHelper)

        if self._full_install or self._java_scripts:
            self._tools.append(self._get_tool(JavaC))

        self._tools.append(self._get_tool(TclLibrary))

        self.tool = self._get_tool(JavaVM)
        self._tools.append(self.tool)

        self._check_tools(self._tools)

    def _add_jar_tool(self, req_tool_class, **kwargs):
        req_tool = self._get_tool(req_tool_class, **kwargs)
        self._tools.append(req_tool)
        self.class_path.append(req_tool.tool_path)

    def prepare(self):
        super(JavaTestRunner, self).prepare()

        self.script = self.get_script_path(required=True)

        self.target_java = str(self.settings.get("compile-target-java", self.target_java))

        self._java_scripts = self._collect_script_files({'.java'})

        self.props_file = self.engine.create_artifact("runner", ".properties")

        self.working_dir = self.engine.create_artifact(self.settings.get("working-dir", "classes"), "")
        if not os.path.exists(self.working_dir):
            os.makedirs(self.working_dir)

        self.reporting_setup(suffix=self.report_file_suffix)

        self.class_path.extend(self.settings.get("additional-classpath", []))
        self.class_path.extend(self.get_scenario().get("additional-classpath", []))

        # expand user-defined class path elements
        self.class_path = [self.engine.find_file(x) for x in self.class_path]

        self._full_install = False  # should we move it to install_required_tools() params?
        self.install_required_tools()

        self._compile_scripts()

    def resource_files(self):
        resources = super(JavaTestRunner, self).resource_files()
        resources.extend(self.get_scenario().get("additional-classpath", []))
        global_additional_classpath = self.settings.get("additional-classpath", [])

        execution_files = self.execution.get('files', [], force_set=True)
        execution_files.extend(global_additional_classpath)  # later we need to fix path for sending into cloud

        return resources

    def _collect_script_files(self, extensions):
        file_list = []
        if os.path.isdir(self.script):
            for root, _, files in os.walk(self.script):
                for test_file in files:
                    if os.path.splitext(test_file)[1].lower() in extensions:
                        path = get_full_path(join(root, test_file))
                        file_list.append(path)
        else:
            if os.path.splitext(self.script)[1].lower() in extensions:
                file_list.append(get_full_path(self.script))
        return file_list

    def _compile_scripts(self):
        """
        Compile .java files
        """
        if not self._java_scripts:
            return

        self.log.debug("Compiling .java files started")

        jar_path = join(self.engine.artifacts_dir, self.working_dir, self.settings.get("jar-name", "compiled.jar"))
        if os.path.exists(jar_path):
            self.log.debug(".java files are already compiled, skipping")
            return

        compile_cl = ["javac",
                      "-source", self.target_java,
                      "-target", self.target_java,
                      "-d", self.working_dir,
                      ]

        compile_cl.extend(["-cp", os.pathsep.join(self.class_path)])
        compile_cl.extend(self._java_scripts)

        with open(self.engine.create_artifact("javac", ".out"), 'ab') as javac_out:
            with open(self.engine.create_artifact("javac", ".err"), 'ab') as javac_err:
                self.log.debug("running javac: %s", compile_cl)
                self.process = shell_exec(compile_cl, stdout=javac_out, stderr=javac_err)
                ret_code = self.process.poll()

                while ret_code is None:
                    self.log.debug("Compiling .java files...")
                    time.sleep(1)
                    ret_code = self.process.poll()

        if ret_code != 0:
            self.log.debug("javac exit code: %s", ret_code)
            with open(javac_err.name) as err_file:
                out = err_file.read()
            raise ToolError("Javac exited with code: %s\n %s" % (ret_code, out.strip()))

        self.log.info("Compiling .java files completed")

        self._make_jar()

    def _make_jar(self):
        """
        move all .class files to compiled.jar
        """
        self.log.debug("Making .jar started")

        with open(join(self.engine.artifacts_dir, "jar.out"), 'ab') as jar_out:
            with open(join(self.engine.artifacts_dir, "jar.err"), 'ab') as jar_err:
                class_files = [java_file for java_file in listdir(self.working_dir) if java_file.endswith(".class")]
                jar_name = self.settings.get("jar-name", "compiled.jar")
                if class_files:
                    compile_jar_cl = ["jar", "-cf", jar_name]
                    compile_jar_cl.extend(class_files)
                else:
                    compile_jar_cl = ["jar", "-cf", jar_name, "."]

                self.process = shell_exec(compile_jar_cl, cwd=self.working_dir, stdout=jar_out, stderr=jar_err)
                ret_code = self.process.poll()

                while ret_code is None:
                    self.log.debug("Making jar file...")
                    time.sleep(1)
                    ret_code = self.process.poll()

        if ret_code != 0:
            with open(jar_err.name) as err_file:
                out = err_file.read()
            raise ToolError("Jar exited with code %s\n%s" % (ret_code, out.strip()))

        self.log.info("Making .jar file completed")


class JUnitTester(JavaTestRunner):
    """
    Allows to test java and jar files
    """
    def install_required_tools(self):
        path = get_full_path(self.settings.get("path"))
        if path:
            if os.path.isfile(get_full_path(path, "")):
                self.log.warning("JUnit path must point to directory.")
                path = get_full_path(path, step_up=1)

            path = os.path.join(path, "{tool_file}")

        config = BetterDict()
        if path:
            config["path"] = path

        self._add_jar_tool(JUnitJupiterApi, config=config)
        self._add_jar_tool(JUnitJupiterEngine, config=config)
        self._add_jar_tool(JUnitPlatformCommons, config=config)
        self._add_jar_tool(JUnitPlatformEngine, config=config)
        self._add_jar_tool(JUnitPlatformLauncher, config=config)
        self._add_jar_tool(JUnitPlatformRunner, config=config)
        self._add_jar_tool(JUnitPlatformSuiteApi, config=config)
        self._add_jar_tool(JUnitVintageEngine, config=config)
        self._add_jar_tool(ApiGuardian, config=config)
        self._add_jar_tool(OpenTest4j, config=config)
        self._add_jar_tool(JUnit, config=config)

        super(JUnitTester, self).install_required_tools()

    def prepare(self):
        self.report_file_suffix = ".ldjson" if self.engine.is_functional_mode() else ".csv"
        super(JUnitTester, self).prepare()
        self.__write_props_file()

    def create_func_reader(self, report_file):
        return FuncSamplesReader(report_file, self.engine, self.log)

    def create_load_reader(self, report_file):
        return JTLReader(report_file, self.log, None)

    def startup(self):
        jar_list = [join(self.working_dir, jar) for jar in listdir(self.working_dir) if jar.endswith(".jar")]
        jar_list.extend(self._collect_script_files({".jar"}))
        self.class_path.extend(jar_list)

        class_path = os.pathsep.join(self.class_path)
        runner_class = "com.blazemeter.taurus.junit.CustomRunner"
        junit_cmd_line = [self.tool.tool_path, "-cp", class_path, "-Djna.nosys=true", runner_class, self.props_file]

        self.process = self._execute(junit_cmd_line)

    def __write_props_file(self):
        def write_prop(name, val):
            if val:
                if isinstance(val, list):
                    val = ",".join(val)

                fds.write("{name}={val}\n".format(name=name, val=val))

        props = get_assembled_value(configs=[self.settings, self.get_scenario(), self.execution], key="properties")
        props = props or BetterDict()

        junit_version = str(self.settings.get("junit-version", "4"))
        if junit_version == "5":
            props.merge({"junit_version": 5})

        with open(self.props_file, 'wt') as fds:
            for key in sorted(props.keys()):
                write_prop(key, props[key])

            fds.write("report_file=%s\n" % self.report_file)

            load = self.get_load()

            write_prop("iterations", load.iterations)
            write_prop("hold_for", load.hold)
            write_prop("concurrency", load.concurrency)
            write_prop("ramp_up", load.ramp_up)
            write_prop("steps", load.steps)

            write_prop("run_items", self._get_items_list("run-items"))
            write_prop("include_category", self._get_items_list("include-categories"))
            write_prop("exclude_category", self._get_items_list("exclude-categories"))

    def _get_items_list(self, param):
        scenario = self.get_scenario()
        scen_items = scenario.get(param, [])
        if isinstance(scen_items, str):
            scen_items = scen_items.split(',')

        exec_items = self.execution.get(param, scen_items)
        if isinstance(exec_items, str):
            exec_items = exec_items.split(',')

        return exec_items


class TestNGTester(JavaTestRunner):
    """
    Allows to test java and jar files with TestNG
    """
    __test__ = False  # Hello, nosetests discovery mechanism

    def install_required_tools(self):
        local_path = self.settings.get("path", None)
        config = {}
        if local_path:
            config = BetterDict.from_dict({"config": local_path})
        self._add_jar_tool(TestNG, config=config)
        super(TestNGTester, self).install_required_tools()

    def _store_testng_xml(self):
        if not self.execution.get('testng-xml') and self.settings.get("autodetect-xml", True):
            script_dir = get_full_path(self.get_script_path(), step_up=1)
            testng_xml = os.path.join(script_dir, 'testng.xml')
            if os.path.exists(testng_xml):
                self.log.info("Detected testng.xml file at %s", testng_xml)
                self.execution['testng-xml'] = testng_xml

    def resource_files(self):
        resources = super(TestNGTester, self).resource_files()
        self._store_testng_xml()
        testng_xml = self.execution.get('testng-xml')
        if testng_xml:
            resources.append(testng_xml)

        return resources

    def startup(self):
        # java -classpath
        # testng.jar:selenium-server.jar:taurus-java-helpers.jar:json.jar:compiled.jar
        # com.blazemeter.taurus.testng.TestNGRunner runner.properties

        jar_list = [join(self.working_dir, jar) for jar in listdir(self.working_dir) if jar.endswith(".jar")]
        jar_list.extend(self._collect_script_files({".jar"}))
        self.class_path.extend(jar_list)
        self._store_testng_xml()

        with open(self.props_file, 'wt') as props:
            props.write("report_file=%s\n" % self.report_file)

            load = self.get_load()
            if load.iterations:
                props.write("iterations=%s\n" % load.iterations)

            if load.hold:
                props.write("hold_for=%s\n" % load.hold)

            for index, item in enumerate(jar_list):
                props.write("target_%s=%s\n" % (index, item.replace(os.path.sep, '/')))

            testng_xml = self.execution.get('testng-xml')
            if testng_xml:
                props.write('testng_config=%s\n' % testng_xml.replace(os.path.sep, '/'))

        cmdline = [self.tool.tool_path, "-cp", os.pathsep.join(self.class_path),
                   "com.blazemeter.taurus.testng.TestNGRunner", self.props_file]
        self.process = self._execute(cmdline)




"""
Aggregating results into DataPoints

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import collections
import copy
import hashlib
import logging
import math
import time
from abc import abstractmethod
from collections import Counter
from typing import Optional, NamedTuple

import fuzzyset
from hdrpy import HdrHistogram, RecordedIterator
from yaml import SafeDumper
from yaml.representer import SafeRepresenter

from bzt import TaurusInternalException, TaurusConfigError
from bzt.engine import Aggregator
from bzt.utils import iteritems, dehumanize_time, JSONConvertible, is_int, to_json

log = logging.getLogger('aggregator')
SAMPLE_STATES = 'success', 'jmeter_errors', 'http_errors'
AGGREGATED_STATES = (
    '_'.join((SAMPLE_STATES[0], SAMPLE_STATES[1])),     # success_jmeter_errors
    '_'.join((SAMPLE_STATES[0], SAMPLE_STATES[2])),     # success_http_errors
    '_'.join((SAMPLE_STATES[2], SAMPLE_STATES[1])),     # http_errors_jmeter_errors
)

ERROR_RESPONSE_BODIES_SIZE_LIMIT: int = 10 * 1024
ERROR_RESPONSE_BODIES_LIMIT: int = 10


class SinglePassIterator(RecordedIterator):
    """
    An iterator to do only one pass on data and calculate all we need inside it.
    Should do one pass and return:
    - stddev
    - percentiles
    - histogram
    """

    def __init__(self, histogram, percentiles, mean):
        super(SinglePassIterator, self).__init__(histogram)
        assert mean is not None, "Known mean is required"
        self.perc_levels = list(percentiles)
        self.perc_levels.sort()
        self._mean = mean
        self._init()

    def reset(self, histogram=None):
        super(SinglePassIterator, self).reset(histogram)
        self._init()

    def _init(self):
        self.percentiles = {}
        self.stdev = 0
        self.hist_values = {}
        self._geometric_dev_total = 0.0
        self._perc_indexes = copy.copy(self.perc_levels)

    def __next__(self):
        item = super(SinglePassIterator, self).__next__()

        # histogram
        self.hist_values[item.value_iterated_to] = item.count_at_value_iterated_to

        # stddev
        dev = (self.histogram._hdr_median_equiv_value(item.value_iterated_to) * 1.0) - self._mean * 1000
        self._geometric_dev_total += (dev * dev) * item.count_added_in_this_iter_step

        # percentiles
        self._fill_percentiles()

        return item

    def _fill_percentiles(self):
        while self._perc_indexes and self._perc_indexes[0] <= self.get_percentile_iterated_to():
            perc_level = self._perc_indexes.pop(0)
            self.percentiles[perc_level] = self.value_at_index

    next = __next__

    def has_next(self):
        has = super(SinglePassIterator, self).has_next()
        if not has:
            self.stdev = math.sqrt(self._geometric_dev_total / self.total_count)
            assert set(self.perc_levels) == set(self.percentiles.keys()), 'Not all percentiles are generated'
        return has


class RespTimesCounter(JSONConvertible):
    def __init__(self, low, high, sign_figures, perc_levels=()):
        super(RespTimesCounter, self).__init__()
        self.low = low
        self.high = high
        self.sign_figures = sign_figures
        self.histogram = HdrHistogram(low, high, sign_figures)
        self._ff_iterator = None
        self._perc_levels = perc_levels
        self.known_mean = None

    def __deepcopy__(self, memo):
        new = RespTimesCounter(self.low, self.high, self.sign_figures)
        new._ff_iterator = self._ff_iterator
        new._perc_levels = self._perc_levels

        new.histogram.counts = copy.deepcopy(self.histogram.counts, memo)
        new.histogram.total_count = self.histogram.total_count
        new.histogram.min_value = self.histogram.min_value
        new.histogram.max_value = self.histogram.max_value

        return new

    def __bool__(self):
        return len(self) > 0

    def __len__(self):
        return self.histogram.total_count

    def add(self, item, count=1):
        item = round(item * 1000.0, 3)
        if item > self.high:
            self.__grow(math.ceil(item / 1000.0) * 1000.0)
        self._ff_iterator = None
        self.histogram.record_value(item, count)

    def merge(self, other):
        self._ff_iterator = None
        if other.high > self.high:
            self.__grow(other.high)

        self.histogram.add(other.histogram)

    def _get_ff(self):
        if self._ff_iterator is None:
            self._ff_iterator = SinglePassIterator(self.histogram, self._perc_levels, self.known_mean)
            for _ in self._ff_iterator:
                pass  # consume it
        return self._ff_iterator

    def get_percentiles_dict(self):
        return self._get_ff().percentiles

    def get_counts(self):
        return self._get_ff().hist_values

    def get_stdev(self):
        return self._get_ff().stdev / 1000.0

    def __json__(self):
        return {
            rt / 1000.0: int(count)  # because hdrpy returns int64, which is unrecognized by json serializer
            for rt, count in iteritems(self.get_counts())
        }

    def __grow(self, newsize):
        log.debug("Growing HDR from %s to %s", self.high, newsize)
        old = self.histogram
        self.high = newsize
        self.histogram = HdrHistogram(self.low, self.high, self.sign_figures)
        self.histogram.add(old)


class ErrorResponseData(NamedTuple):
    content: str
    type: str
    original_size: int

    def get_hash(self) -> str:
        return str(hashlib.md5(to_json(self).encode()).hexdigest())


class KPISet(dict):
    """
    Main entity in results, contains all KPIs for single label,
    capable of merging other KPISet's into it to compose cumulative results
    """
    ERRORS = "errors"
    SAMPLE_COUNT = "throughput"
    CONCURRENCY = "concurrency"
    SUCCESSES = "succ"
    FAILURES = "fail"
    BYTE_COUNT = "bytes"
    RESP_TIMES = "rt"
    AVG_RESP_TIME = "avg_rt"
    STDEV_RESP_TIME = "stdev_rt"
    AVG_LATENCY = "avg_lt"
    AVG_CONN_TIME = "avg_ct"
    PERCENTILES = "perc"
    RESP_CODES = "rc"
    RESP_BODIES = "responseBodies"
    ERRTYPE_ERROR = 0
    ERRTYPE_ASSERT = 1
    ERRTYPE_SUBSAMPLE = 2

    def __init__(self, perc_levels=(), hist_max_rt=1000.0, ext_aggregation=False):
        super(KPISet, self).__init__()
        self.sum_rt = 0
        self.sum_lt = 0
        self.sum_cn = 0
        self.perc_levels = perc_levels
        self.concurrencies = Counter()
        # scalars
        self[KPISet.SAMPLE_COUNT] = 0
        self[KPISet.CONCURRENCY] = 0
        self[KPISet.SUCCESSES] = 0
        self[KPISet.FAILURES] = 0
        self[KPISet.AVG_RESP_TIME] = 0
        self[KPISet.STDEV_RESP_TIME] = 0
        self[KPISet.AVG_LATENCY] = 0
        self[KPISet.AVG_CONN_TIME] = 0
        self[KPISet.BYTE_COUNT] = 0
        # vectors
        self[KPISet.ERRORS] = []
        self[KPISet.RESP_TIMES] = RespTimesCounter(1, hist_max_rt, 3, perc_levels)
        self[KPISet.RESP_CODES] = Counter()
        self[KPISet.PERCENTILES] = {}
        self.ext_aggregation = ext_aggregation

    def __deepcopy__(self, memo):
        mycopy = KPISet(self.perc_levels, self[KPISet.RESP_TIMES].high)
        mycopy.ext_aggregation = self.ext_aggregation
        mycopy.sum_rt = self.sum_rt
        mycopy.sum_lt = self.sum_lt
        mycopy.sum_cn = self.sum_cn
        mycopy.perc_levels = self.perc_levels
        mycopy.concurrencies = copy.deepcopy(self.concurrencies, memo)
        for key in self:
            mycopy[key] = copy.deepcopy(self.get(key, no_recalc=True), memo)
        return mycopy

    @staticmethod
    def error_item_skel(
            error: str, ret_c: str, cnt: int, err_type: int,
            urls: Counter, tag: str, err_resp_data: Optional[ErrorResponseData] = None) -> dict:
        assert isinstance(urls, collections.Counter)
        response_bodies = KPISet._get_response_bodies(err_resp_data)
        return {
            "cnt": cnt,
            "msg": error,
            "tag": tag,  # just one more string qualifier
            "rc": ret_c,
            "type": err_type,
            "urls": urls,
            "responseBodies": response_bodies
        }

    @staticmethod
    def _get_response_bodies(err_resp_data: Optional[ErrorResponseData]) -> list:
        if err_resp_data is None:
            return []

        return [{
            "content": err_resp_data.content,
            "type": err_resp_data.type,
            "original_size": err_resp_data.original_size,
            "hash": err_resp_data.get_hash(),
            "cnt": 1}
        ]

    def add_sample(self, sample):
        """
        Add sample, consisting of: cnc, rt, cn, lt, rc, error, trname, byte_count

        :type sample: tuple
        """
        cnc, r_time, con_time, latency, r_code, error, trname, byte_count = sample
        self[self.SAMPLE_COUNT] += 1
        if cnc:
            self.add_concurrency(cnc, trname)

        if r_code is not None:
            self[self.RESP_CODES][r_code] += 1

            # count times only if we have RCs
            if con_time:
                self.sum_cn += con_time
            self.sum_lt += latency
            self.sum_rt += r_time

        if error is not None:
            self[self.FAILURES] += 1

            item = self.error_item_skel(error, r_code, 1, KPISet.ERRTYPE_ERROR, Counter(), None)
            self.inc_list(self[self.ERRORS], ("msg", error), item)
        else:
            self[self.SUCCESSES] += 1

        self[self.RESP_TIMES].add(r_time, 1)

        if byte_count is not None:
            self[self.BYTE_COUNT] += byte_count

    def add_concurrency(self, cnc, sid):
        # sid: source id, e.g. node id for jmeter distributed mode
        if self.concurrencies.get(sid, 0) < cnc:    # take max value of concurrency during the second.
            self.concurrencies[sid] = cnc

    @staticmethod
    def inc_list(values, selector, value):
        """
        Increment list item, based on selector criteria

        :param values: list to update
        :param selector: tuple of 2 values, field name and value to match
        :param value: dict to put into list
        :type values: list[dict]
        :type selector: tuple
        :type value: dict
        """
        found = False
        for item in values:
            if item[selector[0]] == selector[1]:
                item['cnt'] += value['cnt']
                item['urls'] += value['urls']
                for resp_body in value['responseBodies']:
                    KPISet._inc_resp_body(item, resp_body)
                found = True
                break

        if not found:
            values.append(copy.deepcopy(value))

    @staticmethod
    def _inc_resp_body(item: dict, resp_body: dict) -> None:
        found = False
        for resp_body_data in item['responseBodies']:
            if resp_body_data['hash'] == resp_body['hash']:
                resp_body_data['cnt'] += resp_body['cnt']
                found = True
                break
        if not found:
            item['responseBodies'].append(copy.deepcopy(resp_body))

    def __getitem__(self, key):
        rtimes = self.get(self.RESP_TIMES, no_recalc=True)
        rtimes.known_mean = self.get(self.AVG_RESP_TIME, no_recalc=True)
        if key != self.RESP_TIMES and rtimes:
            if key == self.STDEV_RESP_TIME:
                self[self.STDEV_RESP_TIME] = rtimes.get_stdev()
            elif key == self.PERCENTILES:
                percs = {str(float(perc)): value / 1000.0 for perc, value in
                         iteritems(rtimes.get_percentiles_dict())}
                self[self.PERCENTILES] = percs

        val = super(KPISet, self).__getitem__(key)
        assert key != KPISet.ERRORS or isinstance(val, list)
        return val

    def get(self, k, no_recalc=False):
        if no_recalc:
            return super(KPISet, self).get(k)
        else:
            return self.__getitem__(k)

    def items(self):
        for item in super(KPISet, self).items():
            yield (item[0], self.__getitem__(item[0]))

    def viewvalues(self):
        raise TaurusInternalException("Invalid call")

    def values(self):
        raise TaurusInternalException("Invalid call")

    def recalculate(self):
        """
        Recalculate averages, stdev and percentiles

        :return:
        """
        if self[self.SAMPLE_COUNT]:
            self[self.AVG_CONN_TIME] = self.sum_cn / self[self.SAMPLE_COUNT]
            self[self.AVG_LATENCY] = self.sum_lt / self[self.SAMPLE_COUNT]
            self[self.AVG_RESP_TIME] = self.sum_rt / self[self.SAMPLE_COUNT]

        if len(self.concurrencies):
            self[self.CONCURRENCY] = sum(self.concurrencies.values())

        return self

    def merge_kpis(self, src, sid=None):
        """
        Merge other instance into self

        :param sid: source ID to use when suming up concurrency
        :type src: KPISet
        :return:
        """
        src.recalculate()

        self.sum_cn += src.sum_cn
        self.sum_lt += src.sum_lt
        self.sum_rt += src.sum_rt

        self[self.SAMPLE_COUNT] += src[self.SAMPLE_COUNT]
        self[self.SUCCESSES] += src[self.SUCCESSES]
        self[self.FAILURES] += src[self.FAILURES]
        self[self.BYTE_COUNT] += src[self.BYTE_COUNT]
        # NOTE: should it be average? mind the timestamp gaps
        if src[self.CONCURRENCY]:
            self.add_concurrency(src[self.CONCURRENCY], sid)

        if src[self.RESP_TIMES]:
            self[self.RESP_TIMES].merge(src[self.RESP_TIMES])
        elif not self[self.PERCENTILES]:
            # using existing percentiles, in case we have no source data to recalculate them
            self[self.PERCENTILES] = copy.deepcopy(src[self.PERCENTILES])

        self[self.RESP_CODES].update(src[self.RESP_CODES])

        for src_item in src[self.ERRORS]:
            self.inc_list(self[self.ERRORS], ('msg', src_item['msg']), src_item)

    @staticmethod
    def from_dict(obj, ext_aggregation=False):
        """
        :type obj: dict
        :type ext_aggregation: bool
        :rtype: KPISet
        """
        prc_levels = [float(x) for x in obj[KPISet.PERCENTILES].keys()]
        inst = KPISet(perc_levels=prc_levels, ext_aggregation=ext_aggregation)

        assert inst.PERCENTILES in obj

        for key, val in iteritems(obj):
            if key == inst.RESP_TIMES:
                if isinstance(val, dict):
                    for value, count in iteritems(val):
                        inst[inst.RESP_TIMES].add(float(value), count)
            else:
                inst[key] = val

        inst.sum_cn = obj[inst.AVG_CONN_TIME] * obj[inst.SAMPLE_COUNT]
        inst.sum_lt = obj[inst.AVG_LATENCY] * obj[inst.SAMPLE_COUNT]
        inst.sum_rt = obj[inst.AVG_RESP_TIME] * obj[inst.SAMPLE_COUNT]
        for error in inst[KPISet.ERRORS]:
            error['urls'] = Counter(error['urls'])
        return inst


class DataPoint(dict):
    """
    Represents an aggregate data point

    :param ts: timestamp of this point
    """

    SOURCE_ID = 'id'
    TIMESTAMP = "ts"
    CURRENT = "current"
    CUMULATIVE = "cumulative"
    SUBRESULTS = "subresults"

    def __init__(self, ts, perc_levels=()):
        """

        :type ts: int
        :type perc_levels: list[float]
        """
        super(DataPoint, self).__init__()
        self.perc_levels = perc_levels
        self[self.SOURCE_ID] = None
        self[self.TIMESTAMP] = ts
        self[self.CUMULATIVE] = {}
        self[self.CURRENT] = {}
        self[self.SUBRESULTS] = []

    def __deepcopy__(self, memo):
        new = DataPoint(self[self.TIMESTAMP], self.perc_levels)
        for key in self.keys():
            new[key] = copy.deepcopy(self[key], memo)
        return new

    def __merge_kpis(self, src, dst, sid):
        """
        :param src: dict[str,KPISet]
        :param dst: dict[str,KPISet]
        :param sid: int
        :return:
        """
        for label, val in iteritems(src):
            dest = dst.setdefault(label, KPISet(self.perc_levels, val[KPISet.RESP_TIMES].high))
            if not isinstance(val, KPISet):
                val = KPISet.from_dict(val)
                val.perc_levels = self.perc_levels
            dest.merge_kpis(val, sid)

    def recalculate(self):
        """
        Recalculate all KPISet's
        """
        for val in self[self.CURRENT].values():
            val.recalculate()

        for val in self[self.CUMULATIVE].values():
            val.recalculate()

    def merge_point(self, src, do_recalculate=True):
        """

        :type src: DataPoint
        """
        if self[self.TIMESTAMP] != src[self.TIMESTAMP]:
            msg = "Cannot merge different timestamps (%s and %s)"
            raise TaurusInternalException(msg % (self[self.TIMESTAMP], src[self.TIMESTAMP]))

        self[DataPoint.SUBRESULTS].append(src)

        self.__merge_kpis(src[self.CURRENT], self[self.CURRENT], src[DataPoint.SOURCE_ID])
        self.__merge_kpis(src[self.CUMULATIVE], self[self.CUMULATIVE], src[DataPoint.SOURCE_ID])

        if do_recalculate:
            self.recalculate()


SafeDumper.add_representer(KPISet, SafeRepresenter.represent_dict)
SafeDumper.add_representer(DataPoint, SafeRepresenter.represent_dict)


class ResultsProvider(object):
    """
    :type listeners: list[AggregatorListener]
    """

    def __init__(self):
        super(ResultsProvider, self).__init__()
        self.cumulative = {}
        self.track_percentiles = [0.0, 50.0, 90.0, 95.0, 99.0, 99.9, 100.0]
        self.listeners = []

        self.buffer_scale_idx = None    # string value of the most interesting percentile (e.q. '90.0')
        self.buffer_multiplier = 2      # how many sequential samples we want to get before aggregation (approximately)

        self.buffer_len = 2                     # how many data points we want to collect before aggregation
        self.min_buffer_len = 2                 # small buffer is more responsive but tends to loose data
        self.max_buffer_len = float('inf')      # buffer_len value can be changed during runtime.

        self.histogram_max = 1.0
        self.known_errors = fuzzyset.FuzzySet(use_levenshtein=True)
        self.max_error_count = 100
        self.known_labels = fuzzyset.FuzzySet(use_levenshtein=True)
        self.generalize_labels = 100
        self._redundant_aggregation = False

    def set_aggregation(self, aggregation):
        self._redundant_aggregation = aggregation

    @staticmethod
    def _fuzzy_fold(key, dataset, limit):
        """
        :type key: str
        :type dataset: fuzzyset.FuzzySet
        :type limit: int
        :rtype: str
        """
        if not key or limit <= 0:
            return key

        if not isinstance(key, str):
            key = key.decode('utf-8')

        if key.lower() in dataset.exact_set:
            return key

        size = len(dataset)
        if size >= limit / 4:
            tolerance = (float(size) / float(limit)) ** 2
            threshold = 1 - tolerance
            matches = dataset.get(key)
            if matches:
                for score, result in matches:
                    if score >= threshold:
                        return result
            elif tolerance >= 1.0:
                return next(iter(dataset.exact_set.values()))  # last resort for capping

        dataset.add(key)
        return key

    def _generalize_label(self, label):
        return self._fuzzy_fold(label, self.known_labels, self.generalize_labels)

    def _fold_error(self, error):
        return self._fuzzy_fold(error, self.known_errors, self.max_error_count)

    def add_listener(self, listener):
        """
        Add aggregate results listener

        :type listener: AggregatorListener
        """
        self.listeners.append(listener)

    def __merge_to_cumulative(self, current):
        """
        Merge current KPISet to cumulative
        :param current: KPISet
        """
        for label, data in iteritems(current):
            default = KPISet(
                perc_levels=self.track_percentiles,
                hist_max_rt=data[KPISet.RESP_TIMES].high,
                ext_aggregation=self._redundant_aggregation)
            cumul = self.cumulative.setdefault(label, default)
            cumul.merge_kpis(data)
            cumul.recalculate()

    def datapoints(self, final_pass=False):
        """
        Generator object that returns datapoints from the reader

        :type final_pass: bool
        """
        for datapoint in self._calculate_datapoints(final_pass):
            current = datapoint[DataPoint.CURRENT]
            if datapoint[DataPoint.CUMULATIVE] or not self._ramp_up_exclude():
                self.__merge_to_cumulative(current)
                datapoint[DataPoint.CUMULATIVE] = copy.deepcopy(self.cumulative)
                datapoint.recalculate()

            for listener in self.listeners:
                listener.aggregated_second(datapoint)
            yield datapoint

    @abstractmethod
    def _calculate_datapoints(self, final_pass=False):
        """
        :rtype : list[DataPoint]
        """
        yield

    @abstractmethod
    def _ramp_up_exclude(self):
        """
        :rtype : bool
        """
        return False


class ResultsReader(ResultsProvider):
    """
    Aggregator that reads samples one by one,
    supposed to be attached to every executor
    """
    def __init__(self, perc_levels=None):
        super(ResultsReader, self).__init__()
        self.ignored_labels = []
        self.log = logging.getLogger(self.__class__.__name__)
        self.buffer = {}
        self.min_timestamp = 0      # last aggregated timestamp, older data is obsolete and must be fixed
        if perc_levels is not None:
            self.track_percentiles = perc_levels

    @staticmethod
    def get_mixed_label(label, rc=None, msg=None):
        # it is used for generation of extended label.
        # each label data is split according to sample state (success/error/assert)

        def http_mark():
            http_prefs = ("Embedded resource download error", "Non HTTP response message")
            return any(msg.startswith(http_pref) for http_pref in http_prefs)

        rc = int(rc) if is_int(rc) else 0  # it's jmeter_error by default
        if msg:
            if rc > 299 or http_mark():
                group = SAMPLE_STATES[2]  # http error
            else:
                group = SAMPLE_STATES[1]  # jmeter error
        else:
            group = SAMPLE_STATES[0]  # succeeded sample

        return '-'.join((label, str(group)))

    def __process_readers(self, final_pass=False):
        """

        :param final_pass: True if in post-process stage
        :return:
        """
        for result in self._read(final_pass):
            if result is None:
                self.log.debug("No data from reader")
                break
            elif isinstance(result, list) or isinstance(result, tuple):
                t_stamp, label, conc, r_time, con_time, latency, r_code, error, trname, byte_count = result

                if any([label.startswith(ignore) for ignore in self.ignored_labels]):
                    continue

                if t_stamp < self.min_timestamp:
                    self.log.debug("Putting sample %s into %s", t_stamp, self.min_timestamp)
                    t_stamp = self.min_timestamp

                if r_time < 0:
                    self.log.warning("Negative response time reported by tool, resetting it to zero")
                    r_time = 0

                if t_stamp not in self.buffer:
                    self.buffer[t_stamp] = []

                error = self._fold_error(error)
                self.buffer[t_stamp].append((label, conc, r_time, con_time, latency, r_code, error, trname, byte_count))
            else:
                raise TaurusInternalException("Unsupported results from %s reader: %s" % (self, result))

    def __aggregate_current(self, datapoint, samples):
        """
        :param datapoint: DataPoint
        :param samples: list of samples
        :return:
        """
        current = datapoint[DataPoint.CURRENT]
        for sample in samples:
            # sample format: label, conc, r_time, con_time, latency, r_code, error, trname, byte_count
            base_label = sample[0]

            # empty means overall
            if base_label == '':
                base_label = '[empty]'

            if self.generalize_labels:
                base_label = self._generalize_label(base_label)

            self.__add_sample(current, base_label, sample[1:])

        overall = KPISet(self.track_percentiles, self.__get_rtimes_max(''), ext_aggregation=self._redundant_aggregation)

        for label in current.values():
            overall.merge_kpis(label, datapoint[DataPoint.SOURCE_ID])
        current[''] = overall

        return current

    def _get_suffix(self, label):
        # to collect kpisets to overall sets according to rules result we need to split base label and suffix
        # the suffixes replace '' label in meaning of 'summary result'
        if self._redundant_aggregation:
            return label[label.rfind('-'):]
        return ''

    def __add_sample(self, current, label, kpis):
        if self._redundant_aggregation:
            # kpis format: conc, r_time, con_time, latency, r_code, error_msg, trname, byte_count
            label = self.get_mixed_label(label=label, rc=kpis[4], msg=kpis[5])

        if label not in current:
            current[label] = KPISet(
                perc_levels=self.track_percentiles,
                hist_max_rt=self.__get_rtimes_max(label),
                ext_aggregation=self._redundant_aggregation)

        current[label].add_sample(kpis)

    def __get_rtimes_max(self, label):
        if label in self.cumulative:
            rtimes_max = self.cumulative[label][KPISet.RESP_TIMES].high
        else:
            rtimes_max = self.histogram_max * 1000.0
        return rtimes_max

    def _calculate_datapoints(self, final_pass=False):
        """
        A generator to read available datapoints

        :type final_pass: bool
        :rtype: DataPoint
        """
        if final_pass or len(self.buffer) < self.buffer_len * 10:  # safety valve to preserve RAM
            self.__process_readers(final_pass)
        else:
            self.log.debug("Skipped reading new data, we have enough in the buffer")

        self.log.debug("Buffer len: %s; Known errors count: %s", len(self.buffer), len(self.known_errors))
        if not self.buffer:
            return

        if self.cumulative and self.track_percentiles and self.buffer_scale_idx is not None:
            old_len = self.buffer_len

            # choose average timing of the most interesting percentile
            chosen_timing = self.cumulative[''][KPISet.PERCENTILES][self.buffer_scale_idx]

            # and calculate new buffer_len based on current speed of data getting
            self.buffer_len = round(chosen_timing * self.buffer_multiplier)
            self.buffer_len = max(self.min_buffer_len, self.buffer_len)
            self.buffer_len = min(self.max_buffer_len, self.buffer_len)
            if self.buffer_len != old_len:
                self.log.info("Changed data analysis delay to %ds", self.buffer_len)

        timestamps = sorted(self.buffer.keys())
        while final_pass or (timestamps[-1] >= (timestamps[0] + self.buffer_len)):
            timestamp = timestamps.pop(0)
            self.min_timestamp = timestamp + 1
            self.log.debug("Aggregating: %s, %s in buffer", timestamp, len(self.buffer))
            samples = self.buffer.pop(timestamp)
            datapoint = self.__get_new_datapoint(timestamp)
            self.__aggregate_current(datapoint, samples)
            yield datapoint

            if not timestamps:
                break

    def __get_new_datapoint(self, timestamp):
        """
        :rtype: DataPoint
        """
        point = DataPoint(timestamp, self.track_percentiles)
        point[DataPoint.SOURCE_ID] = id(self)
        return point

    @abstractmethod
    def _read(self, final_pass=False):
        """

        :param final_pass: True if called from post-process stage, when reader
            should report possible rests of results
        :rtype: list
        :return: timestamp, label, concurrency, rt, latency, rc, error
        """
        yield


class ConsolidatingAggregator(Aggregator, ResultsProvider):
    """

    :type underlings: list[bzt.modules.aggregator.ResultsProvider]
    """
    OVERALL_STATE = "all_transactions_aggregated"

    def __init__(self):
        Aggregator.__init__(self, is_functional=False)
        ResultsProvider.__init__(self)
        self.generalize_labels = 500
        self.ignored_labels = ["ignore"]
        self.underlings = []
        self.buffer = {}
        self.histogram_max = 5.0
        self._sticky_concurrencies = {}
        self.min_timestamp = None

        self.collect_error_response_bodies = False
        self.error_response_bodies_limit = ERROR_RESPONSE_BODIES_LIMIT
        self.error_response_bodies_size_limit = ERROR_RESPONSE_BODIES_SIZE_LIMIT

    def converter(self, data):
        if data and self._redundant_aggregation:
            self.__extend_reported_data(data)

        return data

    def __extend_reported_data(self, kpi_sets):
        def add_kpi_set_to_state(destination, _state=self.OVERALL_STATE):
            # add kpi_set to data[<label>][<_state>] value
            if _state not in destination:
                destination[_state] = copy.deepcopy(kpi_set)  # avoid merging kpis for first sample
            else:
                destination[_state].merge_kpis(kpi_set, "")  # deepcopy inside
                destination[_state].recalculate()

        data = kpi_sets['current']
        overall_label = ''
        mixed_labels = set(data.keys()) - {overall_label}
        data[overall_label] = dict()
        for key in mixed_labels:
            sep = key.rindex('-')
            original_label, state = key[:sep], key[sep + 1:]
            kpi_set = data.pop(key)
            if original_label not in data:
                data[original_label] = dict()

            add_kpi_set_to_state(data[overall_label])
            add_kpi_set_to_state(data[original_label])
            add_kpi_set_to_state(data[overall_label], state)
            add_kpi_set_to_state(data[original_label], state)

            for agg_state in AGGREGATED_STATES:
                if state in agg_state:
                    add_kpi_set_to_state(data[overall_label], agg_state)
                    add_kpi_set_to_state(data[original_label], agg_state)
        pass

    def prepare(self):
        """
        Read aggregation options
        """
        super(ConsolidatingAggregator, self).prepare()

        # make unique & sort
        self.track_percentiles = self.settings.get("percentiles", self.track_percentiles)
        self.track_percentiles = list(set(self.track_percentiles))
        self.track_percentiles.sort()
        self.settings["percentiles"] = self.track_percentiles
        
        self.set_aggregation(self.settings.get('extend-aggregation'))

        self.ignored_labels = self.settings.get("ignore-labels", self.ignored_labels)
        self.generalize_labels = self.settings.get("generalize-labels", self.generalize_labels)

        self.collect_error_response_bodies = self.settings.get("collect-error-response-bodies", self.collect_error_response_bodies)
        self.error_response_bodies_limit = self.settings.get("error-response-bodies-limit", self.error_response_bodies_limit)
        self.error_response_bodies_size_limit = self.settings.get("error-response-bodies-size-limit", self.error_response_bodies_size_limit)

        self.min_buffer_len = dehumanize_time(self.settings.get("min-buffer-len", self.min_buffer_len))

        max_buffer_len = self.settings.get("max-buffer-len", self.max_buffer_len)
        try:
            self.max_buffer_len = dehumanize_time(max_buffer_len)
        except TaurusInternalException as exc:
            self.log.debug("Exception in dehumanize_time(%s): %s", max_buffer_len, exc)
            raise TaurusConfigError("Wrong 'max-buffer-len' value: %s" % max_buffer_len)

        self.buffer_multiplier = self.settings.get("buffer-multiplier", self.buffer_multiplier)

        count = len(self.track_percentiles)
        if count == 1:
            self.buffer_scale_idx = str(float(self.track_percentiles[0]))
        if count > 1:
            percentile = self.settings.get("buffer-scale-choice", 0.5)
            percentiles = [i / (count - 1.0) for i in range(count)]
            distances = [abs(percentile - percentiles[i]) for i in range(count)]
            index_position = distances.index(min(distances))
            self.buffer_scale_idx = str(float(self.track_percentiles[index_position]))

        debug_str = 'Buffer scaling setup: percentile %s from %s selected'
        self.log.debug(debug_str, self.buffer_scale_idx, self.track_percentiles)
        self.histogram_max = dehumanize_time(self.settings.get("histogram-initial", self.histogram_max))
        self.max_error_count = self.settings.get("max-error-variety", self.max_error_count)

    def startup(self):
        super(Aggregator, self).startup()

        # send rules to underlings
        for underling in self.underlings:
            underling.set_aggregation(self._redundant_aggregation)

    def add_underling(self, underling):
        """
        Add source for aggregating

        :type underling: ResultsProvider
        """
        underling.track_percentiles = self.track_percentiles
        underling.ignored_labels = self.ignored_labels
        underling.min_buffer_len = self.min_buffer_len
        underling.max_buffer_len = self.max_buffer_len
        underling.buffer_multiplier = self.buffer_multiplier
        underling.buffer_scale_idx = self.buffer_scale_idx
        underling.histogram_max = self.histogram_max

        underling.max_error_count = self.max_error_count
        underling.generalize_labels = self.generalize_labels

        # share error set and label set between underlings
        underling.known_errors = self.known_errors
        underling.known_labels = self.known_labels

        underling.collect_error_response_bodies = self.collect_error_response_bodies
        underling.error_response_bodies_limit = self.error_response_bodies_limit
        underling.error_response_bodies_size_limit = self.error_response_bodies_size_limit

        self.underlings.append(underling)

    def check(self):
        """
        Check if there is next aggregate data present

        :rtype: bool
        """
        for point in self.datapoints():
            self.log.debug("Processed datapoint: %s/%s with %d labels",
                           point[DataPoint.TIMESTAMP], point[DataPoint.SOURCE_ID], len(point[DataPoint.CUMULATIVE]))
        return super(ConsolidatingAggregator, self).check()

    def post_process(self):
        """
        Process all remaining aggregate data
        """
        super(ConsolidatingAggregator, self).post_process()
        for point in self.datapoints(True):
            self.log.debug("Processed datapoint: %s/%s", point[DataPoint.TIMESTAMP], point[DataPoint.SOURCE_ID])

    def _process_underlings(self, final_pass):
        time_start = time.time()
        has_some_time = lambda x: time.time() - x < self.engine.check_interval
        while final_pass or has_some_time(time_start):
            had_data = False
            for underling in self.underlings:
                for point in underling.datapoints(final_pass):
                    had_data = True
                    self._put_into_buffer(point)

            if not had_data:
                break

    def _put_into_buffer(self, point):
        tstamp = point[DataPoint.TIMESTAMP]
        if self.buffer:
            mints = min(self.buffer.keys())
            if tstamp < mints:
                self.log.debug("Putting datapoint %s into %s", tstamp, mints)
                point[DataPoint.TIMESTAMP] = mints
                tstamp = mints
        self.buffer.setdefault(tstamp, []).append(point)

    def _get_max_ramp_up(self):
        ramp_ups = [0]
        for execution in self.engine.config['execution']:
            if 'ramp-up' in execution:
                ramp_ups.append(dehumanize_time(execution['ramp-up']))
        return max(ramp_ups)

    def _ramp_up_exclude(self):
        return self.engine.config.get('settings').get('ramp-up-exclude')

    def _calculate_datapoints(self, final_pass=False):
        """
        Override ResultsProvider._calculate_datapoints
        """
        self._process_underlings(final_pass)

        self.log.debug("Consolidator buffer[%s]: %s", len(self.buffer), self.buffer.keys())
        if not self.buffer:
            return

        timestamps = sorted(self.buffer.keys())
        while timestamps and (final_pass or (timestamps[-1] >= timestamps[0] + self.buffer_len)):
            tstamp = timestamps.pop(0)
            self.log.debug("Merging into %s", tstamp)
            points_to_consolidate = self.buffer.pop(tstamp)

            for subresult in points_to_consolidate:
                if self._ramp_up_exclude():
                    if not self.min_timestamp:
                        self.min_timestamp = subresult['ts']

                    if subresult['ts'] < self.min_timestamp + self._get_max_ramp_up():
                        subresult[DataPoint.CUMULATIVE] = dict()

                if not subresult[DataPoint.SOURCE_ID]:
                    raise ValueError("Reader must provide source ID for datapoint")
                self._sticky_concurrencies[subresult[DataPoint.SOURCE_ID]] = {
                    label: kpiset[KPISet.CONCURRENCY] for label, kpiset in iteritems(subresult[DataPoint.CURRENT])
                }

            if len(points_to_consolidate) == 1:
                self.log.debug("Bypassing consolidation because of single result")
                point = points_to_consolidate[0]
                point[DataPoint.SUBRESULTS] = [points_to_consolidate[0]]
            else:
                point = DataPoint(tstamp, self.track_percentiles)
                for subresult in points_to_consolidate:
                    self.log.debug("Merging %s", subresult[DataPoint.TIMESTAMP])
                    point.merge_point(subresult, do_recalculate=False)
                point.recalculate()

            current_sids = [x[DataPoint.SOURCE_ID] for x in point[DataPoint.SUBRESULTS]]
            for sid in self._sticky_concurrencies:
                if sid not in current_sids:
                    self.log.debug("Adding sticky concurrency for %s", sid)
                    self._add_sticky_concurrency(point, sid)

            point[DataPoint.SOURCE_ID] = self.__class__.__name__ + '@' + str(id(self))
            yield point

    def _add_sticky_concurrency(self, point, sid):
        concur = self._sticky_concurrencies[sid]
        for label, kpiset in iteritems(point[DataPoint.CURRENT]):  # type: (str, KPISet)
            if label in concur:
                kpiset.add_concurrency(concur[label], sid)


class NoneAggregator(Aggregator, ResultsProvider):
    """
    Dummy aggregator
    """

    def __init__(self):
        Aggregator.__init__(self, is_functional=False)
        ResultsProvider.__init__(self)

    def _calculate_datapoints(self, final_pass=False):
        pass


class AggregatorListener(object):
    """
    Mixin for listeners of aggregator data
    """

    @abstractmethod
    def aggregated_second(self, data):
        """
        Notification about new data point

        :param data: bzt.modules.reporting.DataPoint
        """
        pass

    def finalize(self):
        """
        This method is called at the end of run
        to close open file descriptors etc.
        """
        pass



DEV_VERSION = "0.0.0dev0"

try:
    from .version import VERSION
except ImportError:
    VERSION = DEV_VERSION

try:
    from .gitinfo import GIT_INFO
except ImportError:
    GIT_INFO = DEV_VERSION

try:
    from .build import BUILD
except ImportError:
    BUILD = DEV_VERSION



"""
Module holds all stuff regarding usage of Apache Benchmark

Copyright 2016 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import re
from math import ceil
from distutils.version import LooseVersion

from bzt import TaurusConfigError
from bzt.engine import ScenarioExecutor
from bzt.modules.aggregator import ConsolidatingAggregator, ResultsReader
from bzt.modules.console import ExecutorWidget
from bzt.requests_model import HTTPRequest
from bzt.utils import iteritems, CALL_PROBLEMS, shutdown_process, RequiredTool, dehumanize_time, FileReader


class ApacheBenchmarkExecutor(ScenarioExecutor):
    """
    Apache Benchmark executor module
    """

    def __init__(self):
        super(ApacheBenchmarkExecutor, self).__init__()
        self.log = logging.getLogger('')
        self.process = None
        self._tsv_file = None
        self.tool = None
        self.scenario = None

    def prepare(self):
        super(ApacheBenchmarkExecutor, self).prepare()
        self.scenario = self.get_scenario()
        self.install_required_tools()

        self._tsv_file = self.engine.create_artifact("ab", ".tsv")

        self.stdout = open(self.engine.create_artifact("ab", ".out"), 'w')
        self.stderr = open(self.engine.create_artifact("ab", ".err"), 'w')

        self.reader = TSVDataReader(self._tsv_file, self.log)
        if isinstance(self.engine.aggregator, ConsolidatingAggregator):
            self.engine.aggregator.add_underling(self.reader)

    def get_widget(self):
        """
        Add progress widget to console screen sidebar

        :return:
        """
        if not self.widget:
            label = "%s" % self
            self.widget = ExecutorWidget(self, "AB: " + label.split('/')[1])
        return self.widget

    def __first_http_request(self):
        for request in self.scenario.get_requests():
            if isinstance(request, HTTPRequest):
                return request
        return None

    def startup(self):
        args = [self.tool.tool_path]
        load = self.get_load()
        load_iterations = load.iterations or 1
        load_concurrency = load.concurrency or 1

        if load.hold:
            hold = int(ceil(dehumanize_time(load.hold)))
            args += ['-t', str(hold)]
        else:
            args += ['-n', str(load_iterations * load_concurrency)]  # ab waits for total number of iterations

        timeout = self.get_scenario().get("timeout", None)
        if timeout:
            args += ['-s', str(ceil(dehumanize_time(timeout)))]

        args += ['-c', str(load_concurrency)]
        args += ['-d']  # do not print 'Processed *00 requests' every 100 requests or so
        args += ['-r']  # do not crash on socket level errors

        if self.tool.version and LooseVersion(self.tool.version) >= LooseVersion("2.4.7"):
            args += ['-l']  # accept variable-len responses

        args += ['-g', str(self._tsv_file)]  # dump stats to TSV file

        # add global scenario headers
        for key, val in iteritems(self.scenario.get_headers()):
            args += ['-H', "%s: %s" % (key, val)]

        requests = self.scenario.get_requests()
        if not requests:
            raise TaurusConfigError("You must specify at least one request for ab")
        if len(requests) > 1:
            self.log.warning("ab doesn't support multiple requests. Only first one will be used.")
        request = self.__first_http_request()
        if request is None:
            raise TaurusConfigError("ab supports only HTTP requests, while scenario doesn't have any")

        # add request-specific headers
        for key, val in iteritems(request.headers):
            args += ['-H', "%s: %s" % (key, val)]

        if request.method != 'GET':
            raise TaurusConfigError("ab supports only GET requests, but '%s' is found" % request.method)

        if request.priority_option('keepalive', default=True):
            args += ['-k']

        user_cmd = self.settings.get("cmdline")
        if user_cmd:
            args += user_cmd.split(" ")

        args += [request.url]

        self.reader.setup(load_concurrency, request.label)

        self.process = self._execute(args)

    def check(self):
        ret_code = self.process.poll()
        if ret_code is None:
            return False
        if ret_code != 0:
            self.log.warning("ab tool exited with non-zero code: %s", ret_code)
        return True

    def shutdown(self):
        shutdown_process(self.process, self.log)

    def install_required_tools(self):
        self.tool = self._get_tool(ApacheBenchmark, config=self.settings)
        if not self.tool.check_if_installed():
            self.tool.install()

    def get_error_diagnostics(self):
        diagnostics = []
        if self.stdout is not None:
            with open(self.stdout.name) as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("ab STDOUT:\n" + contents)
        if self.stderr is not None:
            with open(self.stderr.name) as fds:
                contents = fds.read().strip()
                if contents.strip():
                    diagnostics.append("ab STDERR:\n" + contents)
        return diagnostics


class TSVDataReader(ResultsReader):
    def __init__(self, filename, parent_logger):
        super(TSVDataReader, self).__init__()
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.file = FileReader(filename=filename, parent_logger=self.log)
        self.skipped_header = False
        self.concurrency = None
        self.url_label = None

    def setup(self, concurrency, url_label):
        self.concurrency = concurrency
        self.url_label = url_label

        return True

    def _read(self, last_pass=False):
        lines = self.file.get_lines(size=1024 * 1024, last_pass=last_pass)

        for line in lines:
            if not self.skipped_header:
                self.skipped_header = True
                continue
            log_vals = [val.strip() for val in line.split('\t')]

            _error = None
            _rstatus = None

            _url = self.url_label
            _concur = self.concurrency
            _tstamp = int(log_vals[1])  # timestamp - moment of request sending
            _con_time = float(log_vals[2]) / 1000.0  # connection time
            _etime = float(log_vals[4]) / 1000.0  # elapsed time
            _latency = float(log_vals[5]) / 1000.0  # latency (aka waittime)
            _bytes = None

            yield _tstamp, _url, _concur, _etime, _con_time, _latency, _rstatus, _error, '', _bytes


class ApacheBenchmark(RequiredTool):
    def __init__(self, config=None, **kwargs):
        settings = config or {}
        tool_path = settings.get('path', 'ab')

        super(ApacheBenchmark, self).__init__(tool_path=tool_path, installable=False, **kwargs)

    def _get_version(self, output):
        version = re.findall("Version\s(\S+)\s", output)
        if not version:
            self.log.warning("%s tool version parsing error: %s", self.tool_name, output)
        else:
            return version[0]

    def check_if_installed(self):
        self.log.debug('Trying %s: %s', self.tool_name, self.tool_path)
        try:
            out, err = self.call([self.tool_path, '-V'])
        except CALL_PROBLEMS as exc:
            self.log.warning("%s check failed: %s", self.tool_name, exc)
            return False

        self.version = self._get_version(out)
        self.log.debug("%s check stdout: %s", self.tool_name, out)
        if err:
            self.log.warning("%s check stderr: %s", self.tool_name, err)
        return True



import ast


def ast_attr(fields):
    """ fields is string of attrs (e.g. 'self.call.me.now') or list of ast args"""
    if isinstance(fields, str):
        if "." in fields:
            fields_list = fields.split(".")
            return ast.Attribute(attr=fields_list[-1], value=ast_attr(".".join(fields_list[:-1])))

        return ast.Name(id=fields)
    else:
        if len(fields) == 1:
            if isinstance(fields[0], str):
                return ast.Name(id=fields[0])
            else:
                return fields[0]

        return ast.Attribute(attr=fields[-1], value=ast_attr(fields[:-1]))  # join ast expressions


def ast_call(func, args=None, keywords=None):
    args = args or []
    if isinstance(func, str):
        func = ast.Name(id=func)
    return ast.Call(func=func, args=args, keywords=keywords or [])


def gen_empty_line_stmt():
    return ast.Expr(value=ast.Name(id=""))  # hacky, but works


def gen_try_except(try_body, exception_body=None, final_body=None):
    if exception_body:
        handlers = [ast.ExceptHandler(
            type=ast.Name(id='Exception', ctx=ast.Load()),
            name=None,
            body=exception_body)]
    else:
        handlers = []

    if not final_body:
        final_body = []

    return ast.Try(
        body=try_body,
        handlers=handlers,
        orelse=None,
        finalbody=final_body,
    )


def gen_raise():
    return ast.Raise(exc=None, cause=None)


def gen_subscript(var_name, index):
    """ Generates code like variable[1]  """
    return ast.Expr(value=ast.Subscript(value=ast.Name(id=var_name, ctx=ast.Load()),
                                        slice=ast.Index(value=ast.Constant(value=index, kind="")), ctx=ast.Load()))


def gen_store(name, value):
    return ast.Assign(
        targets=[ast.Subscript(
            value=ast_attr("self.vars"),
            slice=name)],
        value=value)



"""
Module holds all stuff regarding Siege tool usage

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import time
from math import ceil

from bzt import TaurusConfigError, ToolError
from bzt.engine import ScenarioExecutor
from bzt.modules.aggregator import ConsolidatingAggregator, ResultsReader
from bzt.modules.console import ExecutorWidget
from bzt.requests_model import HTTPRequest
from bzt.utils import iteritems, CALL_PROBLEMS, shutdown_process, RequiredTool, dehumanize_time, FileReader


class SiegeExecutor(ScenarioExecutor):
    def __init__(self):
        super(SiegeExecutor, self).__init__()
        self.log = logging.getLogger('')
        self.process = None
        self.__rc_name = None
        self.__url_name = None
        self.tool = None
        self.scenario = None

    def prepare(self):
        super(SiegeExecutor, self).prepare()
        self.scenario = self.get_scenario()
        self.install_required_tools()

        self.__rc_name = self.execution.get("rc-file")
        if not self.__rc_name:
            config_params = ('verbose = true',
                             'csv = true',
                             'timestamp = false',
                             'fullurl = true',
                             'display-id = true',
                             'show-logfile = false',
                             'logging = false')
            self.__rc_name = self.engine.create_artifact("siegerc", "")
            with open(self.__rc_name, 'w') as rc_artifact_file:
                rc_artifact_file.writelines('\n'.join(config_params))

        self.__url_name = self.get_script_path()

        if self.__url_name:
            self.engine.existing_artifact(self.__url_name)
        elif 'requests' in self.scenario:
            self.__url_name = self._fill_url_file()
        else:
            raise TaurusConfigError("Siege: you must specify either script(url-file) or some requests")

        self.stdout = open(self.engine.create_artifact("siege", ".out"), 'w')
        self.stderr = open(self.engine.create_artifact("siege", ".err"), 'w')

        self.reader = DataLogReader(self.stdout.name, self.log)
        if isinstance(self.engine.aggregator, ConsolidatingAggregator):
            self.engine.aggregator.add_underling(self.reader)

    def resource_files(self):
        script = self.get_script_path()
        if script:
            return [script]
        else:
            return []

    def _fill_url_file(self):
        url_file_name = self.engine.create_artifact("siege", ".url")
        user_vars = self.scenario.get('variables')
        user_vars = ["%s=%s" % (key, val) for (key, val) in iteritems(user_vars)]

        url_list = []
        for req in self.scenario.get_requests():
            if not isinstance(req, HTTPRequest):
                msg = "Siege executor doesn't support '%s' blocks, skipping"
                self.log.warning(msg, req.NAME)
                continue
            url_list.append(req.url)

        with open(url_file_name, 'w') as url_file:
            url_file.writelines('\n'.join(user_vars + url_list))

        return url_file_name

    def startup(self):
        args = [self.tool.tool_path]
        load = self.get_load()

        if load.iterations:
            args += ['--reps', str(load.iterations)]
        elif load.hold:
            hold_for = ceil(dehumanize_time(load.hold))
            args += ['--time', '%sS' % hold_for]
        else:
            raise TaurusConfigError("Siege: You must specify either 'hold-for' or 'iterations'")

        think_time = self.scenario.get_think_time()
        if think_time:
            args += ['--delay', str(dehumanize_time(think_time))]
        else:
            args += ['--benchmark']

        load_concurrency = load.concurrency
        args += ['--concurrent', str(load_concurrency)]
        self.reader.concurrency = load_concurrency

        args += ['--file', self.__url_name]

        for key, val in iteritems(self.scenario.get_headers()):
            args += ['--header', "%s: %s" % (key, val)]

        user_cmd = self.settings.get("cmdline")
        if user_cmd:
            args += user_cmd.split(" ")

        self.env.set({"SIEGERC": self.__rc_name})
        self.process = self._execute(args)

    def check(self):
        ret_code = self.process.poll()
        if ret_code is None:
            return False
        if ret_code != 0:
            raise ToolError("Siege tool exited with non-zero code: %s" % ret_code, self.get_error_diagnostics())
        return True

    def get_widget(self):
        if not self.widget:
            label = "%s" % self
            self.widget = ExecutorWidget(self, "Siege: " + label.split('/')[1])
        return self.widget

    def shutdown(self):
        """
        If tool is still running - let's stop it.
        """
        shutdown_process(self.process, self.log)

    def install_required_tools(self):
        self.tool = self._get_tool(Siege, config=self.settings)
        if not self.tool.check_if_installed():
            self.tool.install()

    def get_error_diagnostics(self):
        diagnostics = []
        if self.stdout is not None:
            with open(self.stdout.name) as fds:
                contents = fds.read().strip()
                if contents:
                    diagnostics.append("Siege STDOUT:\n" + contents)
        if self.stderr is not None:
            with open(self.stderr.name) as fds:
                contents = fds.read().strip()
                if contents:
                    diagnostics.append("Siege STDERR:\n" + contents)
        return diagnostics


class DataLogReader(ResultsReader):
    def __init__(self, filename, parent_logger):
        super(DataLogReader, self).__init__()
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.file = FileReader(filename=filename, parent_logger=self.log)
        self.concurrency = None

    def _read(self, last_pass=False):
        lines = self.file.get_lines(size=1024 * 1024, last_pass=last_pass)

        for line in lines:
            if line.count(chr(0x1b)) != 2:  # skip garbage
                continue
            l_start = line.index('m') + 1
            l_end = line.index(chr(0x1b), l_start)
            line = line[l_start:l_end]
            log_vals = [val.strip() for val in line.split(',')]

            # _mark = log_vals[0]           # 0. current test mark, defined by --mark key
            # _http = log_vals[1]           # 1. http protocol
            _rstatus = log_vals[2]  # 2. response status code
            _etime = float(log_vals[3])  # 3. elapsed time (total time - connection time)
            _rsize = int(log_vals[4])  # 4. size of response
            _url = log_vals[5]  # 6. long or short URL value
            # _url_id = int(log_vals[7])    # 7. url number
            _tstamp = time.strptime(log_vals[7], "%Y-%m-%d %H:%M:%S")
            _tstamp = int(time.mktime(_tstamp))  # 8. moment of request sending

            _con_time = 0
            _latency = 0
            _error = None
            _concur = self.concurrency

            yield _tstamp, _url, _concur, _etime, _con_time, _latency, _rstatus, _error, '', _rsize


class Siege(RequiredTool):
    def __init__(self, config=None, **kwargs):
        settings = config or {}
        tool_path = settings.get("path", "siege")
        super(Siege, self).__init__(tool_path=tool_path, installable=False, **kwargs)

    def check_if_installed(self):
        self.log.debug("Trying %s: %s", self.tool_name, self.tool_path)
        try:
            out, err = self.call([self.tool_path, "-h"])
        except CALL_PROBLEMS as exc:
            self.log.warning("%s check failed: %s", self.tool_name, exc)
            return False

        if err:
            out += err
        self.log.debug("%s output: %s", self.tool_name, out)
        return True



"""
Molotov plugin for recording test results for Taurus. Python 3.5+.

Copyright 2017 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import os
import time

import molotov


report_file = open(os.environ["MOLOTOV_TAURUS_REPORT"], 'w')
samples = dict()
scenarios = dict()


def write_report_item(item):
    report_file.write(json.dumps(item) + "\n")
    report_file.flush()


@molotov.events()
async def handle_request(event, **info):
    if event == 'sending_request':
        request = info['request']
        samples[id(request)] = {
            "ts": time.time(),
            "type": "request",
            "label": str(request.url),
            "elapsed": None,
            "responseCode": None,
            "responseMessage": None,
        }
    elif event == 'response_received':
        response = info['response']
        request = info['request']
        if id(request) in samples:
            sample = samples.pop(id(request))
            sample["elapsed"] = round(time.time() - sample["ts"], 3)
            sample["responseCode"] = str(response.status)
            sample["responseMessage"] = response.reason
            write_report_item(sample)
        else:
            print("WARNING: unmatched response and request: %s %s" % (response, request))


@molotov.events()
async def print_concurrency(event, **info):
    if event == 'current_workers':
        workers = info['workers']
        write_report_item({"ts": time.time(), "type": "workers", "value": workers})


@molotov.events()
async def print_test_case(event, **info):
    if event == 'scenario_start':
        scenario = info['scenario']
        ident = (info['wid'], scenario['name'])
        scenarios[ident] = {"ts": time.time(), "wid": "worker-" + str(info['wid'])}
    elif event == 'scenario_success':
        scenario = info['scenario']
        ident = (info['wid'], scenario['name'])
        if ident not in scenarios:
            print("WARNING: unmatched scenario %s:%s" % (info['wid'], scenario['name']))
            return
        sample = scenarios.pop(ident)
        sample.update({
            "type": "scenario_success",
            "name": scenario['name'],
            "duration": round(time.time() - sample["ts"], 3),
        })
        write_report_item(sample)
    elif event == 'scenario_failure':
        scenario = info['scenario']
        exception = info['exception']
        ident = (info['wid'], scenario['name'])
        if ident not in scenarios:
            print("WARNING: unmatched scenario %s:%s" % (info['wid'], scenario['name']))
            return
        sample = scenarios.pop(ident)
        sample.update({
            "type": "scenario_failure",
            "name": scenario['name'],
            "duration": round(time.time() - sample["ts"], 3),
            "exception": exception.__class__.__name__,
            "errorMessage": str(exception),
        })
        write_report_item(sample)


@molotov.global_teardown()
def close_file():
    if not report_file.closed:
        report_file.close()



"""
Implementations for small services

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import copy
import json
import os
import tempfile
import time
import zipfile
import sys
import shutil

from urllib.request import urlopen
from urllib.error import URLError

from bzt import NormalShutdown, ToolError, TaurusConfigError, TaurusInternalException
from bzt.engine import Service, HavingInstallableTools, Singletone
from bzt.modules.javascript import NPMPackage, NPM
from bzt.modules.jmeter import JMeterExecutor
from bzt.utils import get_stacktrace, communicate, BetterDict, TaurusCalledProcessError, Environment
from bzt.utils import get_full_path, shutdown_process, shell_exec, RequiredTool, is_windows
from bzt.utils import replace_in_config, JavaVM, Node, CALL_PROBLEMS, exec_and_communicate

if not is_windows():
    try:
        from pyvirtualdisplay.smartdisplay import SmartDisplay as Display
    except ImportError:
        from pyvirtualdisplay import Display


class PipInstaller(Service):
    pip_constraints = {
        'setuptools': '79.0.1'
    }

    pip_constraints_file = None

    def __init__(self, packages=None, temp_flag=True):
        super(PipInstaller, self).__init__()
        self.packages = packages or []
        self.versions = BetterDict()
        self.engine = None
        self.temp = temp_flag
        self.target_dir = None
        self.interpreter = sys.executable
        self.pip_cmd = [self.interpreter, "-m", "pip"]
        if not self.pip_constraints_file:
            self.pip_constraints_file = tempfile.NamedTemporaryFile(delete=False, mode='w', newline='', suffix='.txt')
            with open(self.pip_constraints_file.name, 'w') as f:
                for key, value in self.pip_constraints.items():
                    f.write(f'{key}=={value}\n')

    def _check_pip(self):
        cmdline = self.pip_cmd + ["--version"]
        try:
            exec_and_communicate(cmdline)
        except TaurusCalledProcessError as exc:
            self.log.debug(exc)
            raise TaurusInternalException("pip module not found for interpreter %s" % self.interpreter)

    def _get_installed(self):
        cmdline = self.pip_cmd + ["list"]
        out, _ = exec_and_communicate(cmdline)
        out = out.split('\n')[2:-1]
        return dict(zip([line.split(' ')[0] for line in out], [line.strip().split(' ')[-1] for line in out]))

    def _missed(self, packages):
        installed = self._get_installed()
        missed = []
        for package in packages:
            if package not in installed or package in self.versions and installed[package] != self.versions[package]:
                missed.append(package)
        return missed

    def _convert_config_versions(self):
        """
        extract from packages config:
          packages:
            - one
            - two==0.0.0
            - name: three
              version: 0.0.0
        and add to self.packages and self.versions
        """
        packages_list = self.parameters.get("packages", None)
        if not packages_list:
            return

        for package_data in packages_list:
            package, version = None, None
            if isinstance(package_data, dict):
                package, version = package_data['name'], package_data.get("version", None)
            elif isinstance(package_data, str):
                package_params = package_data.split("==")
                package, version = package_params[0], package_params[1] if len(package_params) > 1 else None

            self.packages.append(package)
            if version:
                self.versions[package] = version

    def prepare_pip(self):
        """
        pip-installer expect follow definition:
        - service pip-install
          temp: false   # install to ~/.bzt instead of artifacts dir
          packages:
          - first_pkg
          - second_pkg
        """
        self._check_pip()

        self._convert_config_versions()
        if not self.packages:
            return

        # install into artifacts dir if temp, otherwise into .bzt
        self.temp = self.settings.get("temp", self.temp)
        self.temp = self.parameters.get("temp", self.temp)

        self.target_dir = self.engine.temp_pythonpath if self.temp else self.engine.user_pythonpath

        if not os.path.exists(self.target_dir):
            os.makedirs(get_full_path(self.target_dir), exist_ok=True)

    def prepare(self):
        self.prepare_pip()
        if not self.all_packages_installed():
            self.install()

    def all_packages_installed(self):
        self.packages = self._missed(self.packages)
        self.versions = {package: self.versions[package]
                         for package in self.versions.keys() if package in self.packages}
        return False if self.packages else True

    def install(self):
        if not self.packages:
            self.log.debug("Nothing to install")
            return
        cmdline = self.pip_cmd + ["install", "-t", self.target_dir, "-c", self.pip_constraints_file.name]
        for package in self.packages:
            version = self.versions.get(package, None)
            cmdline += [f"{package}=={version}"] if version else [package]
        self.log.debug("pip-installer cmdline: '%s'" % ' '.join(cmdline))
        try:
            out, err = exec_and_communicate(cmdline)
        except TaurusCalledProcessError as exc:
            self.log.debug(exc)
            for line in exc.output.split('\n'):
                if line.startswith("ERROR"):
                    self.log.error(" ".join(line.split(" ")[1:]))
            return
        if "Successfully installed" in out:
            self.log.info(out.split("\n")[-2])
            for err_line in err.split("\n"):
                if err_line.startswith('WARNING'):
                    self.log.warning(" ".join(err_line.split(" ")[1:]))
                if err_line.startswith('ERROR'):
                    self.log.error(" ".join(err_line.split(" ")[1:]))
        self.log.debug("pip-installer stdout: \n%s" % out)
        if err:
            self.log.debug("pip-installer stderr:\n%s" % err)

    def get_version(self, package):
        installed = self._get_installed()
        return installed[package]

    def post_process(self):
        # might be forbidden on win as tool still work
        if self.packages and self.temp and not is_windows() and os.path.exists(self.target_dir):
            self.log.debug("remove packages: %s" % self.packages)

            shutil.rmtree(self.target_dir)  # it removes all content of directory in reality, not only self.packages


class PythonTool(RequiredTool):
    PACKAGES = []

    def __init__(self, engine, settings, **kwargs):
        tool_path = engine.temp_pythonpath
        version = settings.get("version", None)
        dry_install = settings.get("dry-install", False)
        super(PythonTool, self).__init__(tool_path=tool_path, version=version, dry_install=dry_install, **kwargs)
        self.installer = PipInstaller(temp_flag=True if version else False)
        self.installer.engine = engine
        packages = copy.deepcopy(self.PACKAGES)
        if self.version:
            packages[0] += "==" + self.version
        self.installer.parameters = BetterDict.from_dict({'packages': packages})

    def check_if_installed(self):
        self.log.debug(f"Checking {self.tool_name}.")
        self.installer.prepare_pip()
        result = self.installer.all_packages_installed()
        if not result:
            self.log.warning(f"{self.tool_name} check failed.")
        return result

    def install(self):
        if self.dry_install:
            self.log.info(f"Dry installation for {self.tool_name}")
            return
        self.log.debug(f"Installing {self.tool_name}.")
        self.installer.install()

    def get_version(self):
        return self.installer.get_version(self.PACKAGES[0])

    def post_process(self):
        self.installer.post_process()


class Unpacker(Service):
    UNPACK = 'unpacker'
    FILES = 'files'

    def __init__(self):
        super(Unpacker, self).__init__()
        self.files = []

    def prepare(self):
        packed_list = copy.deepcopy(self.parameters.get(Unpacker.FILES, self.files))
        unpacked_list = []
        for archive in packed_list:
            full_archive_path = self.engine.find_file(archive)
            self.log.debug('Unpacking %s', archive)
            with zipfile.ZipFile(full_archive_path) as zip_file:
                zip_file.extractall(self.engine.artifacts_dir)

            archive = os.path.basename(archive)
            unpacked_list.append(archive[:-4])

        replace_in_config(self.engine.config, packed_list, unpacked_list, log=self.log)


class InstallChecker(Service, Singletone):
    @staticmethod
    def _parse_module_filter(filter_value):
        if isinstance(filter_value, str):
            filter = set(filter_value.strip().split(","))
        elif isinstance(filter_value, (list, dict)):
            filter = set(filter_value)
        else:
            filter = set()
        return filter

    def prepare(self):
        modules = self.engine.config.get("modules")
        problems = []
        include_set = self._parse_module_filter(self.settings.get("include", []))
        exclude_set = self._parse_module_filter(self.settings.get("exclude", []))
        for mod_name in modules:
            if include_set and mod_name not in include_set:
                continue
            if exclude_set and mod_name in exclude_set:
                continue

            try:
                self._check_module(mod_name)
            except KeyboardInterrupt:
                raise  # let the engine handle it
            except BaseException as exc:
                self.log.error("Failed to instantiate module %s", mod_name)
                self.log.debug("%s", get_stacktrace(exc))
                problems.append(mod_name)

        if problems:
            raise ToolError("There were errors while checking for installed tools for %s, see details above" % problems)

        raise NormalShutdown("Done checking for tools installed, will exit now")

    def _check_module(self, mod_name):
        mod = self.engine.instantiate_module(mod_name)

        if not isinstance(mod, HavingInstallableTools):
            self.log.debug("Module %s has no install needs", mod_name)
            return

        self.log.info("Checking installation needs for: %s", mod_name)
        mod.install_required_tools()
        self.log.info("Module is fine: %s", mod_name)


class AndroidEmulatorLoader(Service):
    def __init__(self):
        super(AndroidEmulatorLoader, self).__init__()
        self.emulator_process = None
        self.tool_path = ''
        self.startup_timeout = None
        self.avd = ''
        self.stdout = None
        self.stderr = None

    def prepare(self):
        self.startup_timeout = self.settings.get('timeout', 30)
        config_tool_path = self.settings.get('path', '')
        if config_tool_path:
            self.tool_path = get_full_path(config_tool_path)
            os.environ['ANDROID_HOME'] = get_full_path(config_tool_path, step_up=2)
            self.settings['path'] = self.tool_path
        else:
            # try to read sdk path from env..
            sdk_path = os.environ.get('ANDROID_HOME')
            if sdk_path:
                env_tool_path = os.path.join(sdk_path, 'tools', 'emulator')
                self.tool_path = get_full_path(env_tool_path)
                self.settings['path'] = self.tool_path
            else:
                message = 'Taurus can''t find Android SDK automatically, you must point to emulator with modules.'
                message += 'android-emulator.path config parameter or set ANDROID_HOME environment variable'
                raise TaurusConfigError(message)

        tool = AndroidEmulator(tool_path=self.tool_path, log=self.log)
        if not tool.check_if_installed():
            tool.install()

    def startup(self):
        self.log.debug('Starting android emulator...')
        exc = TaurusConfigError('You must choose an emulator with modules.android-emulator.avd config parameter')
        self.avd = self.settings.get('avd', exc)
        self.stdout = open(os.path.join(self.engine.artifacts_dir, 'emulator-%s.out' % self.avd), 'ab')
        self.stderr = open(os.path.join(self.engine.artifacts_dir, 'emulator-%s.err' % self.avd), 'ab')
        self.emulator_process = shell_exec([self.tool_path, '-avd', self.avd], stdout=self.stdout, stderr=self.stderr)
        start_time = time.time()
        while not self.tool_is_started():
            time.sleep(1)
            if time.time() - start_time > self.startup_timeout:
                raise ToolError("Android emulator %s cannot be loaded" % self.avd)
        self.log.info('Android emulator %s was started successfully', self.avd)

    def tool_is_started(self):
        adb_path = os.path.join(get_full_path(self.tool_path, step_up=2), 'platform-tools', 'adb')
        if not os.path.isfile(adb_path):
            self.log.debug('adb is not found in sdk, trying to use an external one..')
            adb_path = 'adb'
        cmd = [adb_path, "shell", "getprop", "sys.boot_completed"]
        self.log.debug("Trying: %s", cmd)
        try:
            proc = shell_exec(cmd)
            out, _ = communicate(proc)
            return out.strip() == '1'
        except BaseException as exc:
            raise ToolError('Checking if android emulator starts is impossible: %s', exc)

    def shutdown(self):
        if self.emulator_process:
            self.log.debug('Stopping android emulator...')
            shutdown_process(self.emulator_process, self.log)
        if not self.stdout.closed:
            self.stdout.close()
        if not self.stderr.closed:
            self.stderr.close()
        _file = self.stdout.name
        if not os.stat(_file).st_size:
            os.remove(_file)
        _file = self.stderr.name
        if not os.stat(_file).st_size:
            os.remove(_file)


class AppiumLoader(Service):
    def __init__(self):
        super(AppiumLoader, self).__init__()
        self.env = Environment(log=self.log)
        self.appium_process = None
        self.tool_path = ''
        self.tools_dir = "~/.bzt/selenium-taurus/appium-server"
        self.default_path = None
        self.startup_timeout = None
        self.addr = ''
        self.port = ''
        self.stdout = None
        self.stderr = None
        self.appium_python = None
        self.appium_server = None

    def prepare(self):
        self.startup_timeout = self.settings.get('timeout', 30)
        self.addr = self.settings.get('addr', '127.0.0.1')
        self.port = self.settings.get('port', 4723)
        self.tool_path = self.settings.get('path', 'appium')
        self.tools_dir = get_full_path(self.settings.get("tools-dir", self.tools_dir))
        self.default_path = os.path.join(self.tools_dir, "node_modules/.bin/appium")
        self.env.add_path({"NODE_PATH": os.path.join(self.tools_dir, "node_modules")})

        self.install_required_tools()

    def install_required_tools(self):
        node = Node(env=self.env, log=self.log)
        npm = NPM(env=self.env, log=self.log)
        self.appium_python = AppiumPython(engine=self.engine, settings=self.settings, log=self.log)
        self.appium_server = AppiumServer(path=self.tool_path, def_path=self.default_path, tools_dir=self.tools_dir,
                                          node_tool=node, npm_tool=npm)
        required_tools = [node, npm, JavaVM(log=self.log), self.appium_python, self.appium_server]
        for tool in required_tools:
            if not tool.check_if_installed():
                tool.install()

    def startup(self):
        self.log.debug('Starting Appium...')
        self.stdout = open(os.path.join(self.engine.artifacts_dir, 'appium.out'), 'wt')
        self.stderr = open(os.path.join(self.engine.artifacts_dir, 'appium.err'), 'wt')
        self.appium_process = shell_exec([self.appium_server.tool_path, "--log-no-colors"],
                                         stdout=self.stdout, stderr=self.stderr)

        start_time = time.time()
        while not self.tool_is_started():
            time.sleep(1)
            if time.time() - start_time > self.startup_timeout:
                raise ToolError("Appium cannot be loaded")

        self.log.info('Appium was started successfully')

    def tool_is_started(self):
        try:
            response = urlopen("http://%s:%s%s" % (self.addr, self.port, '/wd/hub/sessions'))
            resp_str = response.read()
            if not isinstance(resp_str, str):
                resp_str = resp_str.decode()
            return isinstance(json.loads(resp_str), dict)
        except (URLError, ValueError):
            return False

    def shutdown(self):
        if self.appium_process:
            self.log.debug('Stopping appium...')
            shutdown_process(self.appium_process, self.log)
        if not self.stdout.closed:
            self.stdout.close()
        if not self.stderr.closed:
            self.stderr.close()
        _file = self.stdout.name
        if not os.stat(_file).st_size:
            os.remove(_file)
        _file = self.stderr.name
        if not os.stat(_file).st_size:
            os.remove(_file)

    def post_process(self):
        self.appium_python.post_process()


class AppiumPython(PythonTool):
    PACKAGES = ["Appium-Python-Client"]


class AppiumServer(NPMPackage):
    PACKAGE_NAME = "appium@1.22.0"

    def __init__(self, path=None, def_path=None, **kwargs):
        super(AppiumServer, self).__init__(**kwargs)
        self.tool_path = path
        self.default_path = def_path

    def check_if_installed(self):
        self.log.debug("Trying %s...", self.tool_name)
        if not self._check_path(self.tool_path):
            self.tool_path = self.default_path
            return self._check_path(self.default_path)
        return True

    def _check_path(self, path):
        cmd = [path, '--version']
        try:
            out, err = exec_and_communicate(cmd)
        except CALL_PROBLEMS as exc:
            self.log.debug("Failed to check %s: %s", self.tool_name, exc)
            return False

        if err:
            out += err
        self.log.debug("%s output: %s", self.tool_name, out)
        return True


class AndroidEmulator(RequiredTool):
    def __init__(self, **kwargs):
        super(AndroidEmulator, self).__init__(installable=False, **kwargs)

    def check_if_installed(self):
        self.log.debug("Trying %s...", self.tool_name)

        cmd = [self.tool_path, '-list-avds']
        try:
            output = exec_and_communicate(cmd)
        except CALL_PROBLEMS as exc:
            self.log.debug("Failed to check %s: %s", self.tool_name, exc)
            return False

        self.log.debug("%s output: %s", self.tool_name, output)
        return True


class VirtualDisplay(Service, Singletone):
    """
    Selenium executor
    :type virtual_display: Display
    """

    SHARED_VIRTUAL_DISPLAY = None

    def __init__(self):
        super(VirtualDisplay, self).__init__()
        self.virtual_display = None

    def get_virtual_display(self):
        return self.virtual_display

    def set_virtual_display(self):
        if is_windows():
            self.log.warning("Cannot have virtual display on Windows, ignoring")
            return

        if VirtualDisplay.SHARED_VIRTUAL_DISPLAY:
            self.virtual_display = VirtualDisplay.SHARED_VIRTUAL_DISPLAY
        else:
            width = self.parameters.get("width", 1024)
            height = self.parameters.get("height", 768)
            self.virtual_display = Display(size=(width, height))
            self.virtual_display.start()
            msg = "Starting virtual display[%s] %s"
            self.log.info(msg, (width, height), self.virtual_display.new_display_var)
            VirtualDisplay.SHARED_VIRTUAL_DISPLAY = self.virtual_display

            self.engine.shared_env.set({'DISPLAY': os.environ['DISPLAY']})  # backward compatibility

    def free_virtual_display(self):
        if self.virtual_display and self.virtual_display.is_alive():
            self.virtual_display.stop()
        if VirtualDisplay.SHARED_VIRTUAL_DISPLAY:
            VirtualDisplay.SHARED_VIRTUAL_DISPLAY = None

    def startup(self):
        self.set_virtual_display()

    def check_virtual_display(self):
        if self.virtual_display:
            if not self.virtual_display.is_alive():
                self.log.info("Virtual display out: %s", self.virtual_display.stdout)
                self.log.warning("Virtual display err: %s", self.virtual_display.stderr)
                raise TaurusInternalException("Virtual display failed: %s" % self.virtual_display.return_code)

    def check(self):
        self.check_virtual_display()
        return False

    def shutdown(self):
        self.free_virtual_display()


class JmeterRampup(Service, Singletone):
    def __init__(self):
        super(JmeterRampup, self).__init__()
        self.env = Environment(log=self.log)
        self.rampup_process = None
        self.rampup_addr = None
        self.rampup_port = None
        self.rampup_pass = None
        self.stdout = None
        self.stderr = None

    def prepare(self):
        super(JmeterRampup, self).prepare()

        self.rampup_addr = self.settings.get('rampup_addr', 'localhost')
        self.rampup_port = self.settings.get('rampup_port', 6000)
        self.rampup_pass = self.settings.get('rampup_pass', 'blazemeter')

    def startup(self):
        super(JmeterRampup, self).startup()

        if not self.settings.get('enabled', False):
            self.log.debug('Blazemeter dynamic concurrency control is disabled')
            return

        beanshell_ports = []
        for executor in self.engine.provisioning.executors:
            if not isinstance(executor, JMeterExecutor):
                continue
            try:
                port = executor.__getattribute__("beanshell_port")
                beanshell_ports.append(('localhost', port))
            except AttributeError:
                self.log.warning("No beanshell port for executor: %s", executor)
                continue

        if len(beanshell_ports) == 0:
            self.log.warning('No bsh ports defined in any of jmeter executor.')
            beanshell_ports = [('localhost', 9000)]  # Default one as fallback

        self.log.info(f'Starting Blazemeter dynamic concurrency control... '
                      f'Rampup control: {self.rampup_addr}:{self.rampup_port} '
                      f'Jmeter bsh ports: {beanshell_ports}')
        self.stdout = open(os.path.join(self.engine.artifacts_dir, 'rampup.out'), 'wt')
        self.stderr = open(os.path.join(self.engine.artifacts_dir, 'rampup.err'), 'wt')
        self.rampup_process = shell_exec(['python3',
                                          '-c',
                                          f'from bzt.modules.pyscripts.jmxrampup import JmeterRampupProcess; '
                                          f'jm = JmeterRampupProcess({beanshell_ports}, '
                                          f'"{self.rampup_addr}", {self.rampup_port}, "{self.rampup_pass}", '
                                          f'"{self.engine.artifacts_dir}", {self.log.getEffectiveLevel()}); '
                                          f'jm.run()'],
                                         stdout=self.stdout, stderr=self.stderr)

    def _close_and_remove_empty(self, fd):
        if not fd:
            return

        if not fd.closed:
            fd.close()
        _file = fd.name
        if not os.stat(_file).st_size:
            os.remove(_file)

    def shutdown(self):
        if self.rampup_process:
            self.log.debug('Stopping rampup...')
            shutdown_process(self.rampup_process, self.log)

        self._close_and_remove_empty(self.stdout)
        self._close_and_remove_empty(self.stderr)



""" Monitoring service subsystem """
import os
import re
from logging import Logger

import select
import socket
import subprocess
import time
import traceback
from abc import abstractmethod, ABC
from collections import OrderedDict, namedtuple
from typing import List, Optional

import csv
import psutil
from urllib.parse import urlencode
from urwid import Pile, Text

from bzt import TaurusNetworkError, TaurusInternalException, TaurusConfigError
from bzt.engine import Service, Singletone
from bzt.modules.console import PrioritizedWidget
from bzt.modules.passfail import FailCriterion
from bzt.utils import iteritems, b, stream_decode, dehumanize_time, BetterDict, is_windows


class Monitoring(Service, Singletone):
    """
    :type clients: list[ServerAgentClient]
    :type listeners: list[MonitoringListener]
    """

    def __init__(self):
        super(Monitoring, self).__init__()
        self.listeners = []
        self.clients = []
        self.client_classes = {
            'server-agent': ServerAgentClient,
            'graphite': GraphiteClient,
            'local': LocalClient,
        }

    def add_listener(self, listener):
        assert isinstance(listener, MonitoringListener)
        self.listeners.append(listener)

    def prepare(self):
        super(Monitoring, self).prepare()
        clients = (param for param in self.parameters if param not in ('run-at', 'module'))
        for client_name in clients:
            if client_name in self.client_classes:
                client_class = self.client_classes[client_name]
            else:
                self.log.warning('Unknown monitoring found: %s', client_name)
                continue
            for config in self.parameters.get(client_name, []):
                label = config.get('label', None)

                if client_name == 'local':
                    if any([client for client in self.clients
                            if isinstance(client, self.client_classes[client_name])]):
                        break  # skip the second and following local monitoring clients
                    else:
                        if len(self.parameters.get(client_name, [])) > 1:
                            self.log.warning('LocalMonitoring client found twice, configs will be joined')
                        config = BetterDict()
                        for cfg in self.parameters.get(client_name, []):
                            config.merge(cfg)

                client = client_class(self.log, label, config, self.engine)
                self.clients.append(client)
                client.connect()

    def startup(self):
        for client in self.clients:
            client.start()
        super(Monitoring, self).startup()

    def check(self):
        results = []
        for client in self.clients:
            results.extend(client.get_data())

        if results:
            for listener in self.listeners:
                listener.monitoring_data(results)
        return super(Monitoring, self).check()

    def shutdown(self):
        for client in self.clients:
            client.disconnect()
        super(Monitoring, self).shutdown()

    def get_widget(self):
        widget = MonitoringWidget()
        self.add_listener(widget)
        return widget


class MonitoringListener(object):
    @abstractmethod
    def monitoring_data(self, data: List[dict]):
        pass


class MonitoringClient(object):
    def __init__(self, parent_log, engine):
        self.log = parent_log.getChild(self.__class__.__name__)
        self.engine = engine
        self._last_check = 0  # the last check was long time ago
        self.logs_file = None

    def connect(self):
        pass

    @abstractmethod
    def get_data(self):
        pass

    def disconnect(self):
        pass

    def start(self):
        pass


class LocalClient(MonitoringClient):
    """
    :type monitor: LocalMonitor
    """
    AVAILABLE_METRICS = ['cpu', 'mem', 'disk-space', 'engine-loop', 'bytes-recv',
                         'bytes-sent', 'disk-read', 'disk-write', 'conn-all']

    def __init__(self, parent_log, label, config, engine=None):
        super(LocalClient, self).__init__(parent_log, engine)

        self.config = config

        if not engine:
            self.log.warning('Deprecated constructor detected!')
            self.config['metrics'] = self.AVAILABLE_METRICS  # be generous to old clients

        if label:
            self.label = label
        else:
            self.label = 'local'

        self.monitor = None
        self.interval = None
        self.metrics = None
        self._cached_data = None

    # emulation of deprecated interface
    def engine_resource_stats(self):
        stats = namedtuple("ResourceStats", (
            'cpu', 'disk_usage', 'mem_usage', 'rx', 'tx', 'dru', 'dwu', 'engine_loop', 'conn_all'))

        res_stats = self._get_resource_stats()

        cpu = res_stats.get('cpu', None)
        disk_usage = res_stats.get('disk-space', None)
        mem_usage = res_stats.get('mem', None)
        rx = res_stats.get('bytes-recv', None)
        tx = res_stats.get('bytes-sent', None)
        dru = res_stats.get('disk-read', None)
        dwu = res_stats.get('disk-write', None)
        engine_loop = res_stats.get('engine-loop', None)
        conn_all = res_stats.get('conn-all', None)

        return stats(cpu=cpu, disk_usage=disk_usage, mem_usage=mem_usage, rx=rx, tx=tx,
                     dru=dru, dwu=dwu, engine_loop=engine_loop, conn_all=conn_all)

    def _get_resource_stats(self):
        if not self.monitor:
            raise TaurusInternalException('Local monitor must be instantiated')
        return self.monitor.resource_stats()

    def connect(self):
        exc = TaurusConfigError('Metric is required in Local monitoring client')
        metric_names = self.config.get('metrics', exc)

        bad_list = set(metric_names) - set(self.AVAILABLE_METRICS)
        if bad_list:
            self.log.warning('Wrong metrics found: %s', bad_list)

        good_list = set(metric_names) & set(self.AVAILABLE_METRICS)
        if not good_list:
            raise exc

        self.metrics = list(set(good_list))

        self.monitor = LocalMonitorFactory.create_local_monitor(self.log, self.metrics, self.engine)
        self.log.info(f'Using {self.monitor.__class__.__name__} for engine health monitoring')
        self.interval = dehumanize_time(self.config.get("interval", self.engine.check_interval))

        if self.config.get("logging", False):
            self.logs_file = self.engine.create_artifact("local_monitoring_logs", ".csv")
            with open(self.logs_file, "a", newline='') as mon_logs:
                logs_writer = csv.writer(mon_logs, delimiter=',')
                metrics = ['ts'] + sorted([metric for metric in good_list])
                logs_writer.writerow(metrics)

    def get_data(self):
        now = time.time()

        if now > self._last_check + self.interval:
            self._last_check = now
            self._cached_data = []
            metric_values = self._get_resource_stats()

            if self.logs_file:
                with open(self.logs_file, "a", newline='') as mon_logs:
                    line = [str(round(now))] + [str(metric_values[x]) for x in sorted(metric_values.keys())]
                    logs_writer = csv.writer(mon_logs, delimiter=',')
                    logs_writer.writerow(line)

            for name in self.metrics:
                self._cached_data.append({
                    'source': self.label,
                    'ts': now,
                    name: metric_values[name]})

        return self._cached_data


class BaseLocalMonitor(ABC):
    def __init__(self, parent_logger, metrics: List[str], engine):
        if not engine:
            raise TaurusInternalException('Local monitor requires valid engine instance')
        self._informed_on_mem_issue = False
        self.log = parent_logger.getChild(self.__class__.__name__)
        self.metrics = metrics
        self.engine = engine
        self._disk_counters = None
        self._net_counters = None
        self._last_check = None

    def resource_stats(self):
        if not self._last_check:
            self._disk_counters = self.__get_disk_counters()
            self._net_counters = self.__get_net_counters()
            self._last_check = time.time()
            time.sleep(0.2)  # small enough for human, big enough for machine

        now = time.time()
        interval = now - self._last_check
        self._last_check = now
        return self._calc_resource_stats(interval)

    def _calc_resource_stats(self, interval):
        """
        Get local resource stats

        :return: dict
        """
        result = {}

        if 'mem' in self.metrics:
            result['mem'] = self._get_mem_info()

        if 'disk-space' in self.metrics:
            result['disk-space'] = self.__get_disk_usage(self.engine.artifacts_dir).percent

        if 'engine-loop' in self.metrics:
            result['engine-loop'] = self.engine.engine_loop_utilization

        if 'conn-all' in self.metrics:
            try:
                # take all connections without address resolution
                output = subprocess.check_output(['netstat', '-an'])
                output_lines = stream_decode(output).split('\n')  # in py3 stream has 'bytes' type
                est_lines = [line for line in output_lines if line.find('EST') != -1]
                result['conn-all'] = len(est_lines)
            except BaseException as exc:
                self.log.debug("Failed to get connections info: %s", exc)
                result['conn-all'] = 0

        if 'cpu' in self.metrics:
            result['cpu'] = self._get_cpu_percent()

        if 'bytes-recv' in self.metrics or 'bytes-sent' in self.metrics:
            net = self.__get_net_counters()
            if net is not None:
                tx_bytes = int((net.bytes_sent - self._net_counters.bytes_sent) / float(interval))
                rx_bytes = int((net.bytes_recv - self._net_counters.bytes_recv) / float(interval))
                self._net_counters = net
            else:
                rx_bytes = 0.0
                tx_bytes = 0.0

            if 'bytes-recv' in self.metrics:
                result['bytes-recv'] = rx_bytes
            if 'bytes-sent' in self.metrics:
                result['bytes-sent'] = tx_bytes

        if 'disk-read' in self.metrics or 'disk-write' in self.metrics:
            disk = self.__get_disk_counters()
            if disk is not None:
                dru = int((disk.read_bytes - self._disk_counters.read_bytes) / float(interval))
                dwu = int((disk.write_bytes - self._disk_counters.write_bytes) / float(interval))
                self._disk_counters = disk
            else:
                dru = 0.0
                dwu = 0.0

            if 'disk-read' in self.metrics:
                result['disk-read'] = dru
            if 'disk-write' in self.metrics:
                result['disk-write'] = dwu

        return result

    def __get_disk_counters(self):
        counters = None
        try:
            counters = psutil.disk_io_counters()
        except (RuntimeError, PermissionError) as exc:
            self.log.debug("Failed to get disk io counters: %s", exc)
        if counters is None:
            counters = psutil._common.sdiskio(0, 0, 0, 0, 0, 0)  # pylint: disable=protected-access
            # noinspection PyProtectedMember
        return counters

    def __get_net_counters(self):
        counters = None
        try:
            counters = psutil.net_io_counters()
        except (RuntimeError, PermissionError) as exc:
            self.log.debug("Failed to get net io counters: %s", exc)
        if counters is None:
            counters = psutil._common.snetio(0, 0, 0, 0, 0, 0, 0, 0)  # pylint: disable=protected-access
            # noinspection PyProtectedMember
        return counters

    def __get_disk_usage(self, path):
        disk = None
        try:
            disk = psutil.disk_usage(path)
        except (RuntimeError, PermissionError) as exc:
            self.log.debug("Failed to get disk usage metrics: %s", exc)
        if disk is None:
            disk = psutil._common.sdiskusage(0, 0, 0, 0)  # pylint: disable=protected-access
            # noinspection PyProtectedMember
        return disk

    @abstractmethod
    def _get_mem_info(self):
        pass

    @abstractmethod
    def _get_cpu_percent(self):
        pass


class StandardLocalMonitor(BaseLocalMonitor):

    def __init__(self, parent_logger, metrics, engine):
        super().__init__(parent_logger, metrics, engine)

    def _get_mem_info(self):
        try:
            return psutil.virtual_memory().percent
        except AttributeError:
            if not self._informed_on_mem_issue:
                self.log.debug("Failed to get memory usage: %s", traceback.format_exc())
                self.log.warning("Failed to get memory usage, use -v to get more detailed error info")
                self._informed_on_mem_issue = True

    def _get_cpu_percent(self):
        cpu = None
        try:
            cpu = psutil.cpu_percent()
        except (RuntimeError, PermissionError) as exc:
            self.log.debug("Failed to get cpu percent metrics: %s", exc)
        if cpu is None:
            cpu = 0

        return cpu


class BaseCgroupsLocalMonitor(BaseLocalMonitor, ABC):
    """
    Base class for cgroups and cgroups2 local monitors.
    """

    def __init__(self, cgroup_fs_path: str, parent_logger: Logger, metrics: List[str], engine):
        super().__init__(parent_logger, metrics, engine)
        self._log = parent_logger.getChild(self.__class__.__name__)
        self._cgroup_fs_path = cgroup_fs_path
        # max cpu usage in microseconds per one second
        self._max_cpu_usage = self._get_max_cpu_usage()
        self._max_memory_usage = self._get_max_memory_usage()
        self._cpu_usage1_usec = None
        self._cpu_usage1_time = None

    def _get_max_cpu_usage(self) -> Optional[int]:
        """
        Returns maximum possible cpu usage per one second.
        """
        max_cpu_usage = self._get_cgroups_max_cpu_usage()
        if not max_cpu_usage:
            # if no cgroups limit is in place, then maximum possible cpu usage depends on the number of available cpus
            max_cpu_usage = psutil.cpu_count() * 1000000  # number of cpus * microseconds in one second
        return max_cpu_usage

    @abstractmethod
    def _get_cgroups_max_cpu_usage(self) -> Optional[int]:
        pass

    def _get_max_memory_usage(self) -> Optional[int]:
        max_memory_usage = self._get_cgroups_max_memory_usage()
        if not max_memory_usage:
            # if no cgroups limit is in place, then maximum possible memory usage is total available system memory
            try:
                max_memory_usage = psutil.virtual_memory().total
            except AttributeError:
                self._log.warning(f'Unable to determine total available memory', exc_info=True)
        return max_memory_usage

    @abstractmethod
    def _get_cgroups_max_memory_usage(self) -> Optional[int]:
        pass

    def _get_cpu_percent(self):
        """
        Returns average cpu usage in percentage between the last and this measurement. First call returns 0.
        """
        cpu_delta = None
        total_delta = None
        cpu_usage = 0
        try:
            cpu_usage2_time = time.time()
            cpu_usage2_usec = self._get_cgroups_cpu_usage_snapshot()
            if cpu_usage2_usec and self._cpu_usage1_usec:
                # elapsed cpu time our cgroup consumed in time period between measurements
                cpu_delta = cpu_usage2_usec - self._cpu_usage1_usec
            if self._cpu_usage1_time:
                time_delta = cpu_usage2_time - self._cpu_usage1_time
                # max possible cpu usage per one second adjusted to elapsed time between measurements
                total_delta = self._max_cpu_usage * time_delta
            if cpu_delta and total_delta:
                cpu_usage = round((cpu_delta / total_delta) * 100, 1)
            self._cpu_usage1_usec = cpu_usage2_usec
            self._cpu_usage1_time = cpu_usage2_time
        except BaseException:
            self._log.warning(f'Unable to determine cpu usage', exc_info=True)
        return cpu_usage

    @abstractmethod
    def _get_cgroups_cpu_usage_snapshot(self) -> Optional[int]:
        pass

    def _get_mem_info(self):
        """
        Returns memory usage in percentage of available memory.
        """
        memory_usage_pct = None
        try:
            memory_usage = self._get_cgroups_current_memory_usage()
            if self._max_memory_usage and memory_usage:
                memory_usage_pct = round((memory_usage / self._max_memory_usage) * 100, 1)
        except BaseException:
            self._log.warning(f'Unable to determine memory usage', exc_info=True)
        return memory_usage_pct

    @abstractmethod
    def _get_cgroups_current_memory_usage(self) -> Optional[int]:
        pass

    def _read_int_key_from_cgroups(self, filename: str, semantics: str, key: str) -> Optional[int]:
        path = os.path.join(self._cgroup_fs_path, filename)
        if not os.path.exists(path) or not os.path.isfile(path):
            self._log.warning(f'Unable to determine cgroups {semantics}. File {path} not found')
            return None
        try:
            with open(path) as f:
                lines = f.read().splitlines()
        except IOError:
            self._log.warning(f'Unable to determine cgroups {semantics}. File {path} is not readable')
            return None
        # the file contains multiple lines. Look for line “{key} 80173951“
        return Cgroups2LocalMonitor._get_key_value(lines, key)

    def _read_int_from_cgroups(self, filename: str, semantics: str) -> Optional[int]:
        path = os.path.join(self._cgroup_fs_path, filename)
        if not os.path.exists(path) or not os.path.isfile(path):
            self._log.warning(f'Unable to determine cgroups {semantics}. File {path} not found')
            return None

        try:
            with open(path) as f:
                content = f.readline()
        except IOError:
            self._log.warning(f'Unable to determine cgroups {semantics}. File {path} is not readable')
            return None

        try:
            int_value = int(content)
        except ValueError:
            self._log.debug(f'Unable to determine cgroups {semantics}. Value {content} is not an integer.')
            return None
        # max value can be represented as -1 or max 64bit value
        if int_value < 0 or int_value > 9000000000000000000:
            self._log.debug(f'Unable to determine cgroups {semantics}. Value {int_value} is outside of accepted range.')
            int_value = None
        return int_value

    @staticmethod
    def _get_key_value(lines: List[str], key: str):
        key_value_p = re.compile(r'^([^ ]+)\s([^ ]+)$')
        for line in lines:
            m = key_value_p.match(line)
            if m:
                k = m.group(1)
                if k == key:
                    try:
                        return int(m.group(2))
                    except ValueError:
                        pass
        return None


class Cgroups1LocalMonitor(BaseCgroupsLocalMonitor):
    """
    Local monitor that takes some health metrics from cgroups1 virtual file system for accuracy.
    """
    REQUIRED_FILES = [os.path.join('cpu', 'cpuacct.usage'), os.path.join('memory', 'memory.usage_in_bytes')]

    def __init__(self, cgroup_fs_path: str, parent_logger: Logger, metrics: List[str], engine):
        super().__init__(cgroup_fs_path, parent_logger, metrics, engine)
        self._log = parent_logger.getChild(self.__class__.__name__)

    def _get_cgroups_current_memory_usage(self) -> Optional[int]:
        total_memory_usage = self._read_int_from_cgroups(os.path.join('memory', 'memory.usage_in_bytes'),
                                                         'current memory usage')
        cache_memory_usage = self._read_int_key_from_cgroups(os.path.join('memory', 'memory.stat'), 'cache memory',
                                                             'total_inactive_file')
        if total_memory_usage and cache_memory_usage:
            total_memory_usage = total_memory_usage - cache_memory_usage
        return total_memory_usage

    def _get_cgroups_cpu_usage_snapshot(self) -> Optional[int]:
        cpu_usage = self._read_int_from_cgroups(os.path.join('cpu', 'cpuacct.usage'), 'cpu usage')
        if cpu_usage:
            # cgroups1 cpuacct.usage contains cpu usage in nanoseconds, but we need value in microseconds
            cpu_usage = round(cpu_usage / 1000)
        return cpu_usage

    def _get_cgroups_max_cpu_usage(self) -> Optional[int]:
        """
        Returns maximum possible cpu usage per one second determined from cgroups1 limits. Returns None if there is
        no cgroups limit set.
        """
        quota_us = self._read_int_from_cgroups(os.path.join('cpu', 'cpu.cfs_quota_us'), 'max cpu quota us')
        period_us = self._read_int_from_cgroups(os.path.join('cpu', 'cpu.cfs_period_us'), 'max cpu period us')

        if quota_us and period_us:
            cpu_periods_count = 1 / (period_us / 1000000)
            return int(cpu_periods_count * quota_us)
        self._log.debug(f'No cgroups1 cpu usage limit set: {quota_us} {period_us}')
        return None

    def _get_cgroups_max_memory_usage(self) -> Optional[int]:
        return self._read_int_from_cgroups(os.path.join('memory', 'memory.limit_in_bytes'), 'max memory usage')


class Cgroups2LocalMonitor(BaseCgroupsLocalMonitor):
    """
    Local monitor that takes some health metrics from cgroups2 virtual file system for accuracy.
    """
    REQUIRED_FILES = ['cpu.stat', 'memory.current']

    def __init__(self, cgroup_fs_path: str, parent_logger: Logger, metrics: List[str], engine):
        super().__init__(cgroup_fs_path, parent_logger, metrics, engine)
        self._log = parent_logger.getChild(self.__class__.__name__)

    def _get_cgroups_current_memory_usage(self) -> Optional[int]:
        total_memory_usage = self._read_int_from_cgroups('memory.current', 'current memory usage')
        cache_memory_usage = self._read_int_key_from_cgroups('memory.stat', 'cache memory', 'inactive_file')
        if total_memory_usage and cache_memory_usage:
            total_memory_usage = total_memory_usage - cache_memory_usage
        return total_memory_usage

    def _get_cgroups_cpu_usage_snapshot(self) -> Optional[int]:
        # the file /sys/fs/cgroup/cpu.stat contains multiple lines. Look for line “usage_usec 80173951“
        return self._read_int_key_from_cgroups('cpu.stat', 'cpu usage', 'usage_usec')

    def _get_cgroups_max_cpu_usage(self) -> Optional[int]:
        """
        Returns maximum possible cpu usage per one second determined from cgroups2 limits. Returns None if there is
        no cgroups2 limit set.
        """
        path = os.path.join(self._cgroup_fs_path, 'cpu.max')
        if not os.path.exists(path) or not os.path.isfile(path):
            self._log.warning(f'Unable to determine cgroups2 max cpu usage. File {path} not found')
            return None

        max_cpu_p = re.compile(r'^([^ ]+)\s([^ ]+)$')
        try:
            with open(path) as f:
                content = f.readline()
        except IOError:
            self._log.warning(f'Unable to determine cgroups2 max cpu usage. File {path} is not readable')
            return None
        m = max_cpu_p.match(content)
        if m:
            try:
                quota_us = int(m.group(1))
            except ValueError:
                quota_us = None
            try:
                period_us = int(m.group(2))
            except ValueError:
                period_us = None
            if quota_us and period_us:
                cpu_periods_count = 1 / (period_us / 1000000)
                return int(cpu_periods_count * quota_us)
        self._log.debug(f'No cgroups2 cpu usage limit set: {content}')
        return None

    def _get_cgroups_max_memory_usage(self) -> Optional[int]:
        return self._read_int_from_cgroups('memory.max', 'max memory usage')


class LocalMonitorFactory:

    @classmethod
    def create_local_monitor(cls, parent_logger: Logger, metrics: List[str], engine,
                             mtab_path=None) -> BaseLocalMonitor:
        """
        Constructs concrete instance of BaseLocalMonitor depending on OS and cgroups version.
        """
        log = parent_logger.getChild(cls.__name__)
        if not mtab_path:
            if is_windows():
                return StandardLocalMonitor(parent_logger, metrics, engine)
            mtab_path = '/etc/mtab'
        cgroups_version, cgroups_fs_path = cls._detect_cgroup_info(log, mtab_path)
        if cgroups_version == 2:
            return Cgroups2LocalMonitor(cgroups_fs_path, parent_logger, metrics, engine)
        elif cgroups_version == 1:
            return Cgroups1LocalMonitor(cgroups_fs_path, parent_logger, metrics, engine)
        else:
            return StandardLocalMonitor(parent_logger, metrics, engine)

    @classmethod
    def _detect_cgroup_info(cls, log: Logger, mtab_path: str):
        """
        Detects version of cgroups and base path of cgroups filesystem. By default, it is '/sys/fs/cgroup',
        but it could be customized.
        """
        if not os.path.exists(mtab_path) or not os.path.isfile(mtab_path):
            log.warning(f'Unable to detect cgroups version. File {mtab_path} not found')
            return None, None
        try:
            with open(mtab_path) as f:
                lines = f.read().splitlines()
        except IOError:
            log.warning(f'Unable to detect cgroups version. File {mtab_path} is not readable')
            return None, None

        mtab_p = re.compile(r'^([^ ]+)\s([^ ]+)\s([^ ]+).*$')
        cgroups_version = None
        cgroups_fs_path = None
        for line in lines:
            m = mtab_p.match(line)
            if m:
                fs_type = m.group(3)
                if fs_type == 'cgroup2':
                    base_path = m.group(2)
                    # verify the path contains files we expect to be there, perhaps not all controllers are active
                    files_exist = all([os.path.exists(os.path.join(base_path, required_file)) for required_file in
                                       Cgroups2LocalMonitor.REQUIRED_FILES])
                    if files_exist:
                        return 2, base_path
                elif fs_type == 'cgroup':
                    base_path = os.path.dirname(m.group(2))
                    # verify the path contains files we expect to be there, perhaps not all controllers are active
                    files_exist = all([os.path.exists(os.path.join(base_path, required_file)) for required_file in
                                       Cgroups1LocalMonitor.REQUIRED_FILES])
                    if files_exist:
                        cgroups_version = 1
                        cgroups_fs_path = base_path
        return cgroups_version, cgroups_fs_path


class GraphiteClient(MonitoringClient):
    def __init__(self, parent_log, label, config, engine):
        super(GraphiteClient, self).__init__(parent_log, engine)
        self.config = config
        exc = TaurusConfigError('Graphite client requires address parameter')
        self.address = self.config.get("address", exc)
        self.timeout = int(dehumanize_time(self.config.get("timeout", "5s")))
        self.interval = int(dehumanize_time(self.config.get("interval", "5s")))     # interval for client
        self._cached_data = None
        self.url = self._get_url()

        if label:
            self.host_label = label
        else:
            self.host_label = self.address.replace('http://', '').replace('/', '').replace(':', '_')

        if self.config.get("logging", False):
            self.logs_file = self.engine.create_artifact("Graphite_logs_{}".format(self.host_label), ".csv")
            with open(self.logs_file, "a", newline='') as sa_logs:
                logs_writer = csv.writer(sa_logs, delimiter=',')
                metrics = ['ts'] + [metric for metric in self.config.get("metrics")]
                logs_writer.writerow(metrics)

    def _get_url(self):
        exc = TaurusConfigError('Graphite client requires metrics list')
        params = [('target', field) for field in self.config.get('metrics', exc)]
        from_t = int(dehumanize_time(self.config.get('from', self.interval * 1000)))
        until_t = int(dehumanize_time(self.config.get('until', 0)))
        params += [
            ('from', '-%ss' % from_t),
            ('until', '-%ss' % until_t),
            ('format', 'json')
        ]

        url = self.address + '/render?' + urlencode(params)
        if not url.startswith('http'):
            url = 'http://' + url
        return url

    def _data_transfer(self):
        http_client = self.engine.get_http_client()
        response = http_client.request('GET', self.url, timeout=self.timeout)
        return response.json()

    def _get_response(self):
        try:
            json_list = self._data_transfer()
            msg = "Key 'target' not found in graphite response"
            assert all('target' in dic.keys() for dic in json_list), msg
            return json_list
        except BaseException:
            self.log.debug("Metrics receiving: %s", traceback.format_exc())
            self.log.warning("Fail to receive metrics from %s", self.address)
            return []

    def get_data(self):
        now = time.time()

        if now > self._last_check + self.interval:
            self._cached_data = []
            self._last_check = now
            json_list = self._get_response()
            data_line = [now]

            for element in json_list:
                item = {
                    'ts': now,
                    'source': '%s' % self.host_label}

                for datapoint in reversed(element['datapoints']):
                    if datapoint[0] is not None:
                        item[element['target']] = datapoint[0]
                        data_line.append(datapoint[0])
                        break

                self._cached_data.append(item)

            if self.logs_file:
                with open(self.logs_file, "a", newline='') as g_logs:
                    logs_writer = csv.writer(g_logs, delimiter=',')
                    logs_writer.writerow(data_line)

        return self._cached_data


class ServerAgentClient(MonitoringClient):
    def __init__(self, parent_log, label, config, engine):
        """
        :type parent_log: logging.Logger
        :type config: dict
        """
        super(ServerAgentClient, self).__init__(parent_log, engine)
        self.host_label = label
        exc = TaurusConfigError('ServerAgent client requires address parameter')
        self.address = config.get("address", exc)
        if ':' in self.address:
            self.port = int(self.address[self.address.index(":") + 1:])
            self.address = self.address[:self.address.index(":")]
        else:
            self.port = 4444

        self._partial_buffer = ""
        exc = TaurusConfigError('ServerAgent client requires metrics list')
        metrics = config.get('metrics', exc)

        self._result_fields = [x for x in metrics]

        self._metrics_command = "\t".join([x for x in metrics])
        self.socket = socket.socket()
        self.select = select.select

        self.config = config

        # interval for server (ServerAgent)
        self.interval = int(dehumanize_time(config.get("interval", "1s")))

    def connect(self):
        try:
            self.socket.connect((self.address, self.port))
            self.socket.send(b("test\n"))
            resp = self.socket.recv(4)
            assert resp == b("Yep\n")
            self.log.debug("Connected to serverAgent at %s:%s successfully", self.address, self.port)
        except BaseException as exc:
            self.log.warning("Error during connecting to agent at %s:%s: %s", self.address, self.port, exc)
            msg = "Failed to connect to serverAgent at %s:%s" % (self.address, self.port)
            raise TaurusNetworkError(msg)

        if self.config.get("logging", False):
            self.logs_file = self.engine.create_artifact("SAlogs_{}_{}".format(self.address, self.port), ".csv")
            with open(self.logs_file, "a", newline='') as sa_logs:
                logs_writer = csv.writer(sa_logs, delimiter=',')
                metrics = ['ts'] + sorted([metric for metric in self._result_fields])
                logs_writer.writerow(metrics)

    def disconnect(self):
        self.log.debug("Closing connection with agent at %s:%s...", self.address, self.port)
        try:
            self.socket.send(b("exit\n"))
        except BaseException as exc:
            self.log.warning("Error during disconnecting from agent at %s:%s: %s", self.address, self.port, exc)
        finally:
            self.socket.close()

    def start(self):
        self.socket.send(b("interval:%s\n" % self.interval))
        command = "metrics:%s\n" % self._metrics_command
        self.log.debug("Sending metrics command: %s", command)
        self.socket.send(b(command))
        self.socket.setblocking(False)

    def get_data(self):
        """
        :rtype: list[dict]
        """
        now = time.time()
        readable, writable, errored = self.select([self.socket], [self.socket], [self.socket], 0)
        self.log.debug("Stream states: %s / %s / %s", readable, writable, errored)
        for _ in errored:
            self.log.warning("Failed to get monitoring data from agent at %s:%s", self.address, self.port)

        source = self.host_label if self.host_label else '%s:%s' % (self.address, self.port)

        res = []
        for _sock in readable:
            self._partial_buffer += _sock.recv(1024).decode()
            while "\n" in self._partial_buffer:
                line = self._partial_buffer[:self._partial_buffer.index("\n")]
                self._partial_buffer = self._partial_buffer[self._partial_buffer.index("\n") + 1:]
                self.log.debug("Data line: %s", line)
                values = line.split("\t")
                item = {x: float(values.pop(0)) for x in self._result_fields}
                item['ts'] = now
                item['source'] = source
                res.append(item)

                if self.logs_file:
                    with open(self.logs_file, "a", newline='') as sa_logs:
                        line = [str(round(item['ts']))] + line[:-1].split("\t")
                        logs_writer = csv.writer(sa_logs, delimiter=',')
                        logs_writer.writerow(line)

        return res


class MonitoringWidget(Pile, MonitoringListener, PrioritizedWidget):
    def __init__(self):
        self.host_metrics = OrderedDict()
        self.display = Text("")
        super(MonitoringWidget, self).__init__([self.display])
        PrioritizedWidget.__init__(self, priority=20)

    def monitoring_data(self, data):
        for item in data:
            if item['source'] not in self.host_metrics:
                self.host_metrics[item['source']] = OrderedDict()

            for key in sorted(item.keys()):
                if key not in ("source", "ts"):
                    color = ''
                    if key in self.host_metrics[item['source']]:
                        if self.host_metrics[item['source']][key][0] < item[key]:
                            color = 'warmer'
                        elif self.host_metrics[item['source']][key][0] > item[key]:
                            color = 'colder'

                    self.host_metrics[item['source']][key] = (item[key], color)

        text = []
        for host, metrics in iteritems(self.host_metrics):
            text.append(('stat-hdr', " %s \n" % host))

            if metrics:
                maxwidth = max([len(key) for key in metrics])

                for metric, value in iteritems(metrics):
                    if value[0] is None:
                        rendered = 'N/A'
                    elif isinstance(value[0], int):
                        rendered = "{:,}".format(value[0])
                    elif isinstance(value[0], float):
                        rendered = "{:.3f}".format(value[0])
                    else:
                        rendered = value[0]
                    values = (' ' * (maxwidth - len(metric)), metric, rendered)
                    text.append((value[1], "  %s%s: %s\n" % values))

        self.display.set_text(text)
        self._invalidate()


class MonitoringCriteria(MonitoringListener, FailCriterion):
    def __init__(self, config, owner):
        """
        :type config: bzt.utils.BetterDict
        :type owner: bzt.engine.EngineModule
        """
        super(MonitoringCriteria, self).__init__(config, owner)
        for service in self.owner.engine.services:
            if isinstance(service, Monitoring):
                service.add_listener(self)

    def monitoring_data(self, data):
        for item in data:
            val = self.get_value(item)
            if val is not None:
                self.process_criteria_logic(item['ts'], val)

    def _get_field_functor(self, subject, percentage):
        if '/' not in subject:
            raise TaurusConfigError("Wrong syntax for monitoring criteria subject: %s" % subject)
        host = subject[:subject.index('/')]
        metric = subject[subject.index('/') + 1:]
        return lambda x: (x[metric] if x['source'] == host and metric in x else None)



"""
Module for reporting into http://www.blazemeter.com/ service

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from abc import abstractmethod

import yaml

from bzt import TaurusConfigError
from bzt.bza import Workspace
from bzt.utils import iteritems, BetterDict
from bzt.modules.blazemeter.const import LOC

TAURUS_TEST_TYPE = "taurus"
FUNC_API_TEST_TYPE = "functionalApi"
FUNC_GUI_TEST_TYPE = "functionalGui"


class BaseCloudTest(object):
    """
    :type _user: bzt.bza.User
    :type _project: bzt.bza.Project
    :type _test: bzt.bza.Test
    :type master: bzt.bza.Master
    :type cloud_mode: str
    """

    def __init__(self, user, test, project, test_name, default_location, launch_existing_test, parent_log):
        self.log = parent_log.getChild(self.__class__.__name__)
        self.default_location = default_location
        self._test_name = test_name
        self._last_status = None
        self._sessions = None
        self._started = False
        self._user = user
        self._project = project
        self._test = test
        self.launch_existing_test = launch_existing_test
        self.master = None
        self._workspaces = None
        self.cloud_mode = None
        self.dedicated_ips = False
        self.test_type = None
        self.send_report_email = False

    @abstractmethod
    def prepare_locations(self, executors, engine_config):
        pass

    @abstractmethod
    def resolve_test(self, taurus_config, rfiles, delete_old_files=False):
        pass

    @abstractmethod
    def launch_test(self):
        """launch cloud test"""
        pass

    @abstractmethod
    def sanitize_test(self):
        """Sanitize cloud test"""
        pass

    @abstractmethod
    def start_if_ready(self):
        """start cloud test if all engines are ready"""
        pass

    @abstractmethod
    def get_test_status_text(self):
        pass

    @abstractmethod
    def stop_test(self):
        pass

    def get_master_status(self):
        self._last_status = self.master.get_status()
        return self._last_status


class CloudTaurusTest(BaseCloudTest):
    def prepare_locations(self, executors, engine_config):
        available_locations = {}
        is_taurus4 = True
        workspace = Workspace(self._project, {'id': self._project['workspaceId']})
        for loc in workspace.locations(include_private=is_taurus4):
            available_locations[loc['id']] = loc

        if LOC in engine_config and not is_taurus4:
            self.log.warning("Deprecated test API doesn't support global locations")

        for executor in executors:
            if LOC in executor.execution \
                    and isinstance(executor.execution[LOC], dict):
                exec_locations = executor.execution[LOC]
                self._check_locations(exec_locations, available_locations)
            elif LOC in engine_config and is_taurus4:
                self._check_locations(engine_config[LOC], available_locations)
            else:
                default_loc = self._get_default_location(available_locations)
                executor.execution[LOC] = BetterDict.from_dict({default_loc: 1})

            executor.get_load()  # we need it to resolve load settings into full form

    def _get_default_location(self, available_locations):
        if self.default_location and self.default_location in available_locations:
            return self.default_location

        self.log.debug("Default location %s not found", self.default_location)

        for location_id in sorted(available_locations):
            location = available_locations[location_id]
            if location['sandbox'] and not location.get('purposes', {}).get('functional', False):
                return location_id

        if available_locations:
            location_id = sorted(available_locations.keys())[0]
            self.log.warning("Using first location as default: %s", location_id)
            return location_id

        self.log.warning("List of supported locations for you is: %s", sorted(available_locations.keys()))
        raise TaurusConfigError("No sandbox or default location available, please specify locations manually")

    def _check_locations(self, locations, available_locations):
        for location in locations:
            if location not in available_locations:
                self.log.warning("List of supported locations for you is: %s", sorted(available_locations.keys()))
                raise TaurusConfigError("Invalid location requested: %s" % location)

    def resolve_test(self, taurus_config, rfiles, delete_old_files=False):
        if self.launch_existing_test:
            return

        if self._test is not None:
            test_type = self._test.get("configuration").get("type")
            if test_type != self.test_type:
                self.log.debug("Can't reuse test type %r as Taurus test, will create new one", test_type)
                self._test = None

        if self._test is None:
            test_config = {
                "type": self.test_type,
                "plugins": {
                    "taurus": {
                        "filename": ""  # without this line it does not work
                    }
                }
            }

            self._test = self._project.create_test(self._test_name, test_config)

        if delete_old_files:
            self._test.delete_files()

        taurus_config = yaml.safe_dump(taurus_config, default_flow_style=False, explicit_start=True, canonical=False)
        self._test.upload_files(taurus_config, rfiles)
        self._test.update_props({'configuration': {'executionType': self.cloud_mode}})
        self._test.update_props({
            "shouldSendReportEmail": self.send_report_email
        })

    def sanitize_test(self):
        self._test.update_props({'overrideExecutions': []})
        self._test.update_props({'configuration': {'scriptType': 'taurus'}})

    def launch_test(self):
        self.log.info("Initiating cloud test with %s ...", self._test.address)
        self.master = self._test.start(as_functional=self.test_type in (FUNC_API_TEST_TYPE, FUNC_GUI_TEST_TYPE))
        return self.master.address + '/app/#/masters/%s' % self.master['id']

    def start_if_ready(self):
        self._started = True

    def stop_test(self):
        if self.master:
            self.log.info("Ending cloud test...")
            if not self._last_status:
                self.get_master_status()

            if self._last_status["progress"] >= 100:
                self.master.stop()
            else:
                self.master.terminate()

    def get_test_status_text(self):
        if not self._sessions:
            self._sessions = self.master.sessions()
            if not self._sessions:
                return

        mapping = BetterDict()  # dict(executor -> dict(scenario -> dict(location -> servers count)))
        for session in self._sessions:
            try:
                name_split = [part.strip() for part in session['name'].split('/')]
                location = session['configuration']['location']
                count = session['configuration']['serversCount']
                ex_item = mapping.get(name_split[0], force_set=True)

                if len(name_split) > 1:
                    name = name_split[1]
                else:
                    name = "N/A"

                ex_item.get(name, force_set=True)[location] = count
            except KeyError:
                self._sessions = None

        txt = "%s #%s\n" % (self._test['name'], self.master['id'])
        for executor, scenarios in iteritems(mapping):
            txt += " %s" % executor
            for scenario, locations in iteritems(scenarios):
                txt += " %s:\n" % scenario
                for location, count in iteritems(locations):
                    txt += "  Agents in %s: %s\n" % (location, count)

        return txt
